"""
Version : 1.0 ( 06-22-2022).

DEPENDENCIES:
     -'newton.py' in the folder 'utils'
     -'lambert_W.py' in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np
from proxop.utils.newton import newton_
from proxop.utils.lambert_W import lambert_W


class Jeffrey:
    r"""Compute the proximity operator and the evaluation of gamma*D.

    Where D is the function defined as:


                  /  (x-y) * log(x/y)                 if x > 0 and y > 0
        D(x,y) = |   0                                if x=y=0
                 \   + inf                            otherwise

    'gamma' is the scale factor

            When the inputs are arrays, the outputs are computed element-wise
    INPUTS
    ========
    x          - scalar or ND array
    y          - scalar or ND array with the same size as 'x'
    gamma      - scalar or ND array compatible with the blocks of 'y' [default: gamma=1]
    """

    def __init__(
            self
    ):
        pass

    def prox(
            self,
            x: np.ndarray,
            y: np.ndarray,
            gamma: Union[float, np.ndarray] = 1
    ) -> Tuple[np.ndarray, np.ndarray]:
        self._check(x, gamma)
        if np.size(x) != np.size(y):
            raise Exception("'x' and 'y' must have the same size")
        # scalar-like inputs handling
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        if np.size(y) <= 1:
            y = np.reshape(y, (-1))

        # 2nd branch
        sz = np.shape(x)
        prox_p = np.zeros(sz)
        prox_q = np.zeros(sz)

        # branch selection
        # when the input is too big, we use a taylor approximation of the
        # LambertW function to avoid divergence
        u1 = 1 - x / gamma
        u2 = 1 - y / gamma
        lw_x, lw_y = np.zeros(sz), np.zeros(sz)
        # With x ...
        mask_u = u1 > 100
        lw_x[mask_u] = u1[mask_u] - u1[mask_u] / (1 + u1[mask_u]) * np.log(u1[mask_u])

        # we use the Lambert_W function otherwise
        mask_u = np.logical_not(mask_u)
        lw_x[mask_u] = lambert_W(np.exp(1 - x / gamma))
        # With y ...
        mask_u = u2 > 100
        lw_y[mask_u] = u2[mask_u] - u2[mask_u] / (1 + u2[mask_u]) * np.log(u2[mask_u])

        # we use the Lambert_W function otherwise
        mask_u = np.logical_not(mask_u)
        lw_y[mask_u] = lambert_W(np.exp(1 - y / gamma))

        mask = lw_x * lw_y < 1
        gg = gamma
        if np.size(gamma) > 1:
            gg = gamma[mask]
        xx = x[mask]
        yy = y[mask]

        # newton's method

        def fun_phi(u):
            return (
                    (u + 1) * np.log(u) - 1 / u + u ** 2 + (xx / gg - 1) * u + 1
                    - yy / gg
            )

        def der_fun_phi(u):
            return np.log(u) + 1 / u + 1 / u ** 2 + 2 * u + xx / gg

        # root finding
        eps = 1e-10
        low = lw_x + eps
        high = 1 / (lw_y + eps)
        t = (low + high) / 2
        t = newton_(fun_phi, fp=der_fun_phi, x0=t, low=low, high=high)

        # 1st branch
        prox_p[mask] = xx + gg * (np.log(t) + t - 1)
        prox_q[mask] = yy - gg * (np.log(t) - 1 / t + 1)

        return tuple([prox_p, prox_q])

    def __call__(self, x: np.ndarray, y: np.ndarray) -> float:
        if np.size(x) != np.size(y):
            raise Exception("'x' and 'y' must have the same size")
        if (
                np.any(x < 0)
                or np.any(y < 0)
                or np.any((x == 0) * (y != 0))
                or np.any((y == 0) * (x != 0))
        ):
            return np.inf
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
            y = np.reshape(y, (-1))
        mask = y > 0
        res = (x[mask] - y[mask]) * np.log(x[mask] / y[mask])
        return np.sum(res)

    def _check(self, x, gamma=1):
        if np.any(gamma <= 0):
            raise Exception("'gamma'  must be strictly positive")
        if np.size(gamma) > 1 and np.size(gamma) != np.size(x):
            ValueError(
                "'gamma' must be positive scalars or positive ND arrays" +
                " with the same size as 'x'"
            )



"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.newton import newton_


class LogInverse:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is 'logarithm+inverse' function defined as:

                      / ( - log(x) + w * x^(-1) )     if  x> 0
               f(x)= |
                     \  +INF                                  otherwise

        with w > 0 and q >= 1

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>LogInverse()(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'. So, the command >>>LogInverse().prox(x)   will return an
    array with the same shape as 'x'.


     INPUTS
    ========
     x     - scalar or ND array
     w     - positive , scalar or ND array with the same size as 'x'
     gamma - positive, scalar or ND array with the same size as 'x' (default: gamma=1)

    =======
    Examples
    ========

     Evaluate the function  f:

     >>> LogInverse()( np.e )
     -0.6321205588285

      Compute the result element-wise sum for vector inputs :

     >>> LogInverse(w=[1, 2, 3])( [1, 3, np.e] )
     0.6716927015128838

     Compute the proximity operator at a given point :

     >>> LogInverse(w=[1,2,3]).prox(  [-2, 3, 4 ])
     array([0.41421356, 3.30277564, 4.23606798])

     Use a scale factor 'gamma'>0 to compute the proximity operator of  the
     function 'gamma*f'

     >>> LogInverse( [1,2,3,3]).prox( [-2, 3, 4, np.e ], gamma=2.5)
     array([1.3218996 , 3.95255089, 4.83733009, 3.86652778])
    """

    def __init__(self, w: Union[float, np.ndarray] = 1):
        if np.size(w) > 1 and (not isinstance(w, np.ndarray)):
            w = np.array(w)
        if np.any(w < 0):
            raise ValueError(
                "'w (or all its components if it is an array) must be positive"
            )
        self.w = w

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check(x, gamma)
        w = self.w

        def polynom_phi(t):
            return t**3 - x * t**2 - gamma * t - w * gamma

        def der_phi(t):
            return 3 * t**2 - 2 * x * t - gamma

        prox_x = 2*np.abs(x)
        prox_x = newton_(polynom_phi, fp=der_phi, x0=prox_x, low=0, high=np.inf)
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        self._check(x, 1)
        result = np.zeros_like(1.0 * x)
        mask = x > 0
        ww = self.w
        if np.size(self.w) > 1:
            ww = self.w[mask]
        result[mask] = -np.log(x[mask]) + ww * 1 / x[mask]
        result[x <= 0] = np.inf
        return np.sum(result)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it is an array) must "
                + "be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")
        if (np.size(self.w) > 1) and (np.size(self.w) != np.size(x)):
            raise ValueError("'w' must be either scalar or the same size as 'x'")



"""
Created on Mon Jun 13 2022.

author: mbaye diongue
"""

import numpy as np


def non_linear_sys(x, gamma, axis=None):
    """Find the scalar s, solution of the system defined below.

    sum_{n=1}^{n=N}( max(0, x))= lam

    Parameters
    ----------
    x : ND array

    gamma : scalar

    axis: direction of block-wise processing
    Returns
    -------
    gamma : scalar
    """
    # 0. Linearize
    sz = np.size(x)
    if axis is None:
        x = x[:]

    # 1. Order the column elements (sort in decreasing order)
    s = -np.sort(-x, axis=axis)

    # 2. Compute the partial sums: c(j) = ( sum(s(1:j)) - gamma ) / j

    c = (np.cumsum(s, axis=axis) - gamma) / np.cumsum(
        np.ones(np.size(s)), axis=axis
    )

    # 3. Find the index: n = max{ j \in {1,...,B} : s(j) > c(j) }

    mask = np.where(s > c)[0]
    n = np.max(mask, axis)

    # 4. Compute the prox
    p = x - c[n]

    # 5. Output is zero if 's(j) > c(j)' for all j
    p = p * (1 - np.all(mask, axis))

    # 6. revert back
    p = np.reshape(p, sz)

    return p



"""
Version : 1.0 ( 06-18-2022).

Author  : Mbaye DIONGUE

DEPENDENCIES:
   -'newton.py' in the folder 'utils'

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.newton import newton_


class BurgCauchy:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

      where the function f is defined as:


                       /   -log(x)+ w*log( delta + x^2 )   if x>0
               f(x) = |
                      \   +inf                      otherwise

                with w>0 and delta>0

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x    - ND array
    gamma- positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    delta- positive, scalar or ND array with the same size as 'x'[default: delata=1e-10]
    w    - positive, scalar or ND array with the same size as 'x' [default: w=1]

    Note: When calling the function (not the proximity operator) the result
    is computed element-wise SUM. So the command >>>BurgCauchy(delta=0.01, w=2)(x) will
    return a scalar even if x is a vector:

    >>> BurgCauchy(delta=0.01, w=2)(np.array([10, 20, 30]))
    26.098816

    But as expected, >>>BurgCauchy(delta=0.01, w=2).prox(x)
    will return a vector with the same size as x:

    >>> BurgCauchy(delta=0.01, w=2).prox(np.array([10, 20, 30]))
    array([ 9.69046116, 19.84886296, 29.89966593])
    """

    def __init__(
        self,
        w: Union[np.ndarray, float] = 1.0,
        delta: Union[np.ndarray, float] = 1e-10
    ):
        if np.any(delta <= 0):
            raise ValueError(
                "'delta'(or all of its elements)" + " must be strictly positive"
            )
        if np.any(w <= 0):
            raise ValueError(
                "'w' (or all of its elements)" + " must be strictly positive"
            )
        self.delta = delta
        self.w = w

    def prox(self, x: np.ndarray, gamma: Union[np.ndarray, float] = 1.) -> np.ndarray:
        self._check(x, gamma)
        delta = self.delta
        w = self.w
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        def f(t):
            return (
                t**4
                - x * t**3
                + (delta + 2 * gamma * w - gamma) * t**2
                - delta * x * t
                - delta * gamma
            )

        def df(t):
            return (
                4 * t**3
                - 3 * x * t**2
                + 2 * (delta + 2 * gamma * x - gamma) * t
                - delta * x
            )

        abs_x = np.abs(x)
        prox_x = np.maximum(1, abs_x)
        low = 0
        prox_x = newton_(f, df, prox_x, low=low, high=np.inf)
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.any(x <= 0):
            return np.inf
        return np.sum(-np.log(x) + self.w * np.log(self.delta + x**2))

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma'(or all of its elements)" + " must be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("'gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-22-2022).

DEPENDENCIES:
     - 'newton.py' located in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Optional, Tuple
import numpy as np
from proxop.utils.newton import newton_


class PersSqrt:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:


                /  -sqrt(xi^2 - ||y||_2^2)       if xi > 0 and ||y||_2^2 <= xi
      f(y,xi) = |  0                             if y = 0 and xi = 0
                \  +inf                          otherwise


    'gamma' is the scale factor

    INPUTS
    ========
     y         - scalar or ND array
     xi        - scalar or ND array compatible with the blocks of 'y'
     gamma     - positive, scalar or ND array compatible with the blocks of 'x'
                 [default: gamma=1]
     axis      - None or int, axis of block-wise processing [default: axis=None]
                  axis = None --> 'x' is processed as a single vector [DEFAULT] In this
                  case 'gamma' must be a scalar.
                  axis >=0   --> 'x' is processed block-wise along the specified axis
                  (0 -> rows, 1-> columns ect). In this case 'gamma' and 'xi' must be
                  singletons along 'axis'.
    """

    def __init__(
            self,
            xi: Union[float, np.ndarray] = 1,
            gamma: Union[float, np.ndarray] = 1,
            axis: Optional[int] = None
    ):
        if np.any(gamma <= 0):
            raise Exception("'gamma'  must be strictly positive")

        if axis is None and np.size(gamma) > 1:
            raise Exception(
                "An 'axis' must be specified when 'gamma' is not a scalar"
            )
        if (axis is not None) and (axis < 0):
            axis = None
        if np.size(gamma) <= 1:
            gamma = np.reshape(gamma, (-1))
        if np.size(xi) <= 1:
            xi = np.reshape(xi, (- 1))

        self.axis = axis
        self.gamma = gamma
        self.xi = xi

    def prox(self, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        self._check(y)
        scale = self.gamma

        axis = self.axis
        xi = self.xi
        sz0 = np.shape(y)
        if np.size(y) <= 1:
            y = np.array([y])

        sz = np.shape(y)
        sz = np.array(sz, dtype=int)
        sz[axis] = 1

        if np.size(scale) > 1:
            scale = np.reshape(scale, sz)
        if np.size(xi) > 1:
            xi = np.reshape(xi, sz)

        # linearize
        if axis is None:
            y = np.reshape(y, (-1))

        # initialization
        prox_q = np.zeros(sz)

        # condition
        yy = np.linalg.norm(y, ord=2, axis=axis).reshape(sz)
        mask = xi + np.sqrt(scale**2 + yy) > 0
        if np.size(scale) > 1:
            scale = scale[mask]
        if np.size(xi) > 1:
            xi = xi[mask]

        yy_mask = yy[mask]

        # compute t
        def fun_phi(u):
            return 2 * scale * u + xi * u / np.sqrt(1 + u ** 2) - yy_mask

        def der_fun_phi(u):
            return 2 * scale + (
                    xi * np.sqrt(1 + u ** 2) - xi * u ** 2 / np.sqrt(1 + u ** 2)
            ) / (1 + u ** 2)

        # Finding a root with the Newton method
        t = newton_(fun_phi, fp=der_fun_phi, x0=yy_mask, low=0, high=np.inf)

        factor = np.ones_like(yy)
        factor[mask] = scale * t / yy_mask
        prox_p = (y - y * factor)*mask
        qq = xi - scale * np.sqrt(1 + t**2)
        prox_q[mask] = qq
        # revert back
        prox_p = np.reshape(prox_p, sz0)
        return tuple([prox_p, prox_q])

    def __call__(self, y: np.ndarray) -> float:
        if np.size(y) <= 1:
            y = np.array([y])
        xi = self.xi
        # evaluate the function
        yy = np.sum(y**2, self.axis)
        if np.any(xi < 0) or np.any(yy > xi**2):
            return np.inf

        p = -np.sqrt(xi**2 - yy)
        return np.sum(self.gamma * p)

    def _check(self, x):
        sz = np.shape(x)
        if len(sz) <= 1:
            self.axis = None
        if len(sz) <= 1 and (np.size(self.gamma) > 1 or np.size(self.xi) > 1):
            raise Exception(
                "'gamma' and 'xi' must be scalar when 'x' is one dimensional"
            )

        if len(sz) > 1 and (self.axis is not None):
            sz = np.array(sz, dtype=int)
            sz[self.axis] = 1
            if np.size(self.gamma) > 1 and (
                np.prod(sz) != np.size(self.gamma)
            ):
                raise Exception(
                    "The dimension of 'gamma' is not compatible with the blocks of 'x'"
                )
            if np.size(self.xi) > 1 and (np.prod(sz) != np.size(self.xi)):
                raise Exception(
                    "The dimension of 'xi is not compatible with the blocks of 'x'"
                )



"""
Version : 1.0 ( 06-13-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Optional
import numpy as np


class AffineBarrier:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the indicator of an affine barrier defined as:

                               / -log(b- a.T*x)    if u.T*x < b
                        f(x) =|
                              \   + inf            otherwise

    'gamma' is the scale factor


    As f is an indicator funtion, its proximity operator is the projection onto
    the affine barrier set.

     INPUTS
    ========
     x         - ND array
     a         - ND array with the same shape as 'x'
     b         - scalar or ND array compatible with the blocks of 'x'
     gamma     - positive, scalar or ND array compatible with the blocks of 'x'
             (default: gamma=1)
     axis     - None or int, axis of block-wise processing [default: axis=None]
              axis = None --> 'x' is processed as a single vector [DEFAULT] In this case
             'gamma' must be a scalar
              axis >=0   --> 'x' is processed block-wise along the specified axis
              (0 -> rows, 1-> columns ect. in this case, 'gamma' and 'epsilon'
              must be singletons along 'axis')

     =========
     Examples
     ==========


     >>> import numpy as np
     >>> from proxop import AffineBarrier
     >>> x=np.array([1,2,3])
     >>> a= np.array([-1, 5, 3])
     >>> b= 3.5
     >>> AffineBarrier(x,b)(x)
     inf

    As the result below is infinite, that means 'x' does not belong to the affine set.

    Projection of x onto the affine set:

     >>> px= AffineBarrier(x,b).prox(x)
     >>> px
     >>> AffineBarrier(x,b)(px)
     -0.17973361

    As we could have suspected, the projection of 'x' belongs to the affine set.

     >>> x=np.arange(6)
     >>> x=x.reshape((2,3))
     >>> x
     array([[0, 1, 2],
           [3, 4, 5]])
     >>> a =np.ones
     >>> a[0,:]=2
     >>> a
     array([[2., 2., 2.],
            [1., 1., 1.]])
     >>> b=np.array([-1, 2, 4])

     Set 'axis=0' to process along the rows of the matrix 'x':

     >>> AffineBarrier(x,b, axis=0)(x)
     inf

     Projection of x onton the affine set:

     >>> px = AffineBarrier(x,b, axis=0).prox(x)
     >>> px
     array([[0.        , 0.05537521, 0.19926499],
            [0.        , 0.22150085, 0.49816248]])
     >>> AffineBarrier(x,b, axis=0)(px)
     array([-0.        , -0.05696748, -0.10495226])
    """

    def __init__(
            self,
            a: np.ndarray,
            b: np.ndarray,
            axis: Optional[int] = None
    ):
        if (axis is not None) and (axis < 0):
            axis = None
        self.axis = axis
        self.a = a
        self.b = b

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        self._check(x, gamma)
        scale = gamma
        a = self.a
        b = self.b
        sz = np.shape(x)
        sz = np.array(sz, dtype=int)
        sz[self.axis] = 1
        if np.size(scale) > 1:
            scale = np.reshape(scale, sz)
        if np.size(b) > 1:
            b = np.reshape(b, sz)

        scalar_prod = np.sum(self.a * x, axis=self.axis).reshape(sz)
        l2_a = np.sum(self.a ** 2, axis=self.axis).reshape(sz)
        return (
                x
                + (b - scalar_prod - np.sqrt((b - scalar_prod) ** 2 + 4 * scale * l2_a))
                / (2 * l2_a)
                * a
        )

    def __call__(self, x: np.ndarray) -> float:
        scalar_prod = np.sum(self.a * x, self.axis)
        b = self.b
        if np.size(b) != np.size(scalar_prod):
            raise ValueError("'b' must be compatible with the blocks of 'x'")
        b = np.reshape(b, np.shape(scalar_prod))
        res = np.zeros_like(1.0 * scalar_prod)
        mask = scalar_prod < self.b
        res[mask] = -np.log(b[mask] - scalar_prod[mask])
        res[np.logical_not(mask)] = np.inf
        return np.sum(res)

    def _check(self, x, gamma=1):
        scale = gamma
        if np.size(scale) <= 1:
            return

        if np.any(gamma <= 0):
            raise Exception(
                "'gamma' (or all of its components if it is an array)" +
                " must be strictly positive"
            )
        if self.axis is None and np.size(gamma) > 1:
            raise ValueError(
                "'gamma' must be a scalar when the parameter 'axis' is equal to 'None'"
            )
        sz = np.shape(x)
        if len(sz) <= 1:
            self.axis = None

        if len(sz) <= 1:
            raise ValueError(
                "'gamma' must be scalar when 'x' is one dimensional"
            )

        if len(sz) > 1 and (self.axis is not None):
            sz = np.array(sz, dtype=int)
            sz[self.axis] = 1
            if np.size(scale) > 1 and (np.prod(sz) != np.size(scale)):
                raise ValueError(
                    "The dimension of 'gamma' is not compatible" +
                    " with the blocks of 'x'"
                )



"""
Compute the proximity operator and the evaluation of 'scalar' functions.

Note: When the input is a vector, the proximity operator is computed element-wise
(i.e. component by component) thanks to the property of proximity operator.

Author: Mbaye DIONGUE (Jun 2022)
"""


from .AbsValue import *
from .ArctanActi import *
from .ArgsinhActi import *
from .Bathtub import *
from .BentIdentity import *
from .Berhu import *
from .BregAbsEntropy import *
from .BregAbsLog import *
from .BregBoltzShannon import *
from .BregBoltzShannon2 import *
from .BregExp import *
from .BregLpNorm import *
from .BregSquareLog import *
from .ElliotActi import *
from .ELUacti import *
from .Entropy import *
from .Exp import *
from .FairPotential import *
from .FermiDiracEntropy import *
from .GMActi import *
from .HingeLoss import *
from .Huber import *
from .Hyperbolic import *
from .Inverse import *
from .ISRLU import *
from .ISRU import *
from .Log import *
from .LogActi import *
from .LogBarrier import *
from .LogInverse import *
from .LogisticLoss import *
from .LogPower import *
from .NegativeRoot import *
from .Power import *
from .PRelu import *
from .SmoothedFermiDiracEntropy import *
from .Square import *
from .SquaredBathtub import *
from .SquaredHinge import *
from .TanhActi import *
from .Thresholder import *
from .UnimodalSigmoid import *

__all__ = [
    "AbsValue",
    "ArctanActi",
    "ArgsinhActi",
    "Bathtub",
    "BentIdentity",
    "Berhu",
    "BregAbsEntropy",
    "BregAbsLog",
    "BregBoltzShannon",
    "BregBoltzShannon2",
    "BregExp",
    "BregLpNorm",
    "BregSquareLog",
    "ElliotActi",
    "ELUacti",
    "Entropy",
    "Exp",
    "FairPotential",
    "FermiDiracEntropy",
    "GMActi",
    "HingeLoss",
    "Huber",
    "Hyperbolic",
    "Inverse",
    "ISRLU",
    "ISRU",
    "Log",
    "LogActi",
    "LogBarrier",
    "LogInverse",
    "LogisticLoss",
    "LogPower",
    "NegativeRoot",
    "Power",
    "PRelu",
    "SmoothedFermiDiracEntropy",
    "Square",
    "SquaredBathtub",
    "SquaredHinge",
    "TanhActi",
    "Thresholder",
    "UnimodalSigmoid",
]



"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class Log:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is defined as:

                      /  - log(x)          if  x> 0
               f(x)= |
                     \    +INF              otherwise


    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>Log()(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'. So, the command >>>Log().prox(x)   will
    return an array with the same shape as 'x'

     INPUTS
    ========
     x     - scalar or ND array
     gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]

    =======
    Examples
    ========

     Evaluate the function  f:

     >>> Log()( np.e )
     -1.0

      Compute the resuslt as an element-wise sum when the input is a vector input:

     >>> Log()( [1, 3, np.e] )
      -2.09861228866811

     Compute the proximity operator at a given point :

     >>> Log().prox(  [-2, 3, 4 ])
     array([0.41421356, 3.30277564, 4.23606798])

     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function
     'gamma*f'

     >>> Log().prox( [-2, 3, 4, np.e ], gamma=2.5)
     array([0.87082869, 3.67944947, 4.54950976, 3.44415027])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check(x, gamma)
        return 0.5 * (x + np.sqrt(x**2 + 4 * gamma))

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        result = np.zeros(np.shape(x))
        mask = x > 0
        result[mask] = - np.log(x[mask])
        result[x <= 0] = np.inf
        return np.sum(result)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it is an array)"
                + " must be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 08-09-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Optional
import numpy as np


class SoftmaxActi:
    r"""Compute the proximity operator of the softmax activation function.

    The softmax activation function is defined as:

               / \sum_{k=1}^{N}( x_k log(x_k) - x_k^2/2 ) if (x_1,...,x_N) \in [0,1]^N
         f(x)=|                                           and \sum_{k=1}^{N} x_k =1
              \  + inf                                    otherwise

    Note: The proximity operator of this function f is the (shifted) softmax function
    defined as:

        softmax(x) = ( exp(x_k) / ( \sum_{k=1}^{N}exp(x_k) )_{1<=k<=N) - u

    where u = (1,...,1).T a vector which have all components equal to 1

     INPUTS
    ========
     x         - ND array

     axis    - None or int, axis of block-wise processing [default: axis=None]
                  axis = None --> 'x' is processed as a single vector [DEFAULT] In this
                  case 'gamma' must be a scalar.
                  axis >=0   --> 'x' is processed block-wise along the specified axis
                  (0 -> rows, 1-> columns ect. In this case, 'gamma' must be singleton
                  along 'axis'.

     tol   - float, default: tol=1e-10
            Precision tolerance (for example if |x_i| < tol x_i is assumed
            to be null)
     """

    def __init__(self, axis: Optional[int] = None, tol: float = 1e-10):
        self.axis = axis
        self.tol = tol

    def prox(self, x: np.ndarray) -> np.ndarray:
        exp_x = np.exp(x - np.max(x))  # for numerical stability purpose
        shape_x = np.shape(x)
        shape_sum = np.array(shape_x, dtype=int)
        shape_sum[self.axis] = 1
        prox_x = exp_x / exp_x.sum(axis=self.axis).reshape(shape_sum) - np.ones(shape_x)
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        def fun_phi(t):
            if np.any(t > 1) or np.any(t < 0) or \
                    np.any(np.abs(np.sum(t, self.axis) - 1) > self.tol):
                return np.inf
            phi_t = np.zeros(np.shape(t))
            mask = t > 0
            phi_t[mask] = t[mask] * np.log(t[mask])
            return np.sum(phi_t - 0.5 * t ** 2, axis=self.axis)

        fun_x = fun_phi(x + 1.0) + np.sum(x, self.axis)
        return np.sum(fun_x)



"""
Version : 1.0 ( 06-08-2022).

DEPENDENCIES :
     lambert_W.py - located in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.lambert_W import lambert_W


class Exp:
    r"""Computes the proximity operator and the evaluation of gamma*f.

    Where f is the exponential function:

            f(x)= exp(x)


    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>Exp()(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'. So, the command >>>Exp().prox(x)   will
    return an array with the same shape as 'x'

     INPUTS
    ========
    x      -  array_like
    gamma  - positive, scalar or ND array with the same size as 'x'
            [default value: gamma=1]

    =======
    Examples
    ========

     Evaluate the 'direct' function :

     >>> Exp()( np.log(2) )
     2.0

     Compute the result element-wise Sum for vector inputs
     >>> Exp(gamma=2)( [-1, 0, 3, np.log(2)] )
      46.90683272871822

     Compute the proximity operator at a given point :

     >>> Exp().prox(  [-2, 3, 4, np.log(2)] )
     array([-2.12002824,  0.79205997,  1.07372894, -0.15945832])

     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function
     'gamma*f':

     >>> Exp().prox( [-2, 3, 4, np.log(3) ], gamma=2.5 )
      array([-2.26069497,  0.13594702,  0.37230635, -0.46761867])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check(x, gamma)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        prox_x = np.zeros(np.shape(x))
        # we use a taylor approximation to avoid divergence when the input is 'too big'
        u = np.log(gamma) + x
        mask = u > 100
        u = u[mask]
        prox_x[mask] = u - u / (1 + u) * np.log(u)
        not_mask = np.logical_not(mask)
        if np.size(gamma) > 1:
            gamma = gamma[not_mask]
        prox_x[not_mask] = lambert_W(gamma * np.exp(x[not_mask]))
        return x - prox_x

    def __call__(self, x: np.ndarray) -> float:
        return np.sum(np.exp(x))

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its elements if it is an array)"
                + "must be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 (06-07-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class Huber:
    r"""Compute the proximity operator and evaluate of the gamma*f.

    Where f is the Huber Loss function definded as :


                  /  1/2 * x^2             if |x| <= w
           f(x)= |
                 \  w* |x| - w^2  /2       otherwise


                 with w>0


    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>Huber()(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'. So, the command >>>Huber.prox(x)   will return an array
    with the same shape as 'x'.

    ========
     x      - ND array
     w      - positive, scalar or ND array with the same size as 'x'
     gamma  - positive, scalar or ND array with the same size as 'x'[default: gamma=1]

    =======
    Examples
    ========

     Evaluate the 'direct' function :

     >>> Huber(2)( 0.25 )
      0.03125

      Compute the result element-wise for vector inputs :

     >>> Huber(2)( [-1, 3, 0.25] )
     4.53125

     Compute the proximity operator at a given point :

     >>> Huber(3).prox(  [-2, 3, 4 ])
      array([-1,  1,  2])

     Use a scale factor 'gamma'>0 to compute the proximity operator of  the
     function 'gamma*f'

     >>> Huber(3).prox( [-2, 3, 4], gamm2.5)
     array([-0.57142857,  0.85714286,  1.14285714])
    """

    def __init__(self, w: Union[float, np.ndarray]):
        if np.size(w) > 1 and (not isinstance(w, np.ndarray)):
            w = np.array(w)
        if np.any(w <= 0):
            raise Exception(
                "'w' (or all of its components if it is an array) must"
                + " be strictly positive"
            )
        self.w = w

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check(x, gamma)
        w = self.w

        # preliminaries
        abs_x = np.abs(x)
        sign_x = np.sign(x)

        # 2nd branch
        prox_x = x - gamma * w * sign_x

        # 1st branch
        mask = abs_x <= w * (gamma + 1)
        gamma_ = gamma
        if np.size(gamma) > 1:
            gamma_ = gamma[mask]
        prox_x[mask] = x[mask] / (1 + gamma_)
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        abs_x = np.abs(x)
        w = self.w
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            if abs_x <= w:
                return np.sum((0.5 * x**2))
            return np.sum((w * abs_x - w**2 / 2))

        fun_x = 0.5 * x**2
        mask = abs_x > w
        if np.size(w) > 1:
            fun_x[mask] = abs_x[mask] * w[mask] - w[mask] ** 2 / 2
        else:
            fun_x[mask] = abs_x[mask] * w - w**2 / 2
        return np.sum(fun_x)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its elements if it is an array)"
                + "must be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise Exception("gamma' must be either scalar or the same size as 'x'")
        if (np.size(self.w) > 1) and (np.size(self.w) != np.size(x)):
            raise Exception("w' must be either scalar or the same size as 'x'")



"""
Created on Mon Jun 13 2022.

@author: mbaye diongue
"""
import numpy as np


def pava_python(x):
    """Compute the proximity operator of  the monotone Cone.

       S(j,k,:) = sum( x(j:k,:) ) / (k-j+1)

    for every 'j' and 'k' in {1,...,N}, where N = size(x,1).
    """
    N = np.size(x, 0)

    res = np.zeros(np.size(x))
    cum_sum = np.reshape(1 + np.arange(N**2) % N, (N, N))
    cum_sum = 1 + cum_sum - np.arange(1, N+1).reshape((N, 1))
    cum_sum = np.cumsum(x) / cum_sum
    for n in range(N):
        max_sum = -np.inf
        for j in range(n+1):
            min_sum = np.min( cum_sum, axis=1) 
            if min_sum > max_sum:
                max_sum = min_sum
        res[n] = max_sum
    return res



"""
Version : 1.0 ( 06-22-2022).

DEPENDENCIES:
    -'newton.py'  - located in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np
from proxop.utils.newton import newton_


class Kullback:
    r"""Compute the proximity operator and the evaluation of gamma*D.

    Where D is the function defined as:


                  /  x * log(x/y)                     if x > 0 and y > 0
        D(x,y) = |   0                                if x=y=0
                 \   + inf                            otherwise

    'gamma' is the scale factor

            When the inputs are arrays, the outputs are computed element-wise
    INPUTS
    ========
    x          - scalar or ND array
    y          - scalar or ND array with the same size as 'x'
    gamma      - scalar or ND array compatible with the blocks of 'y'[default: gamma=1]
    """

    def __init__(
            self
    ):
        pass

    def prox(
            self,
            x: np.ndarray,
            y: np.ndarray,
            gamma: Union[float, np.ndarray] = 1
    ) -> Tuple[np.ndarray, np.ndarray]:
        self._check(x, gamma)
        if np.size(x) != np.size(y):
            raise Exception("'x' and 'y' must have the same size")
        # scalar-like inputs handling
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
            y = np.reshape(y, (-1))

        # 2nd branch
        sz = np.shape(x)
        prox_p = np.zeros(sz)
        prox_q = np.zeros(sz)

        # branch selection
        mask = np.exp(x / gamma - 1) >= -y / gamma
        xx = x[mask]
        yy = y[mask]
        gg = gamma
        if np.size(gg) > 1:
            gg = gamma[mask]

        # newton's method
        def fun_phi(t):
            return t * np.log(t) + (xx / gg - 1) * t - 1 / t - yy / gg

        def der_fun_phi(t):
            return np.log(t) + xx / gg + 1 / t ** 2

        # root finding
        root = np.exp(-xx / gg)
        low = 1 / np.e * root
        root = newton_(fun_phi, fp=der_fun_phi, x0=root, low=low, high=np.inf)

        # 1st branch
        prox_p[mask] = xx + gg * np.log(root) - gg
        prox_q[mask] = yy + gg / root

        return tuple([prox_p, prox_q])

    def __call__(
            self,
            x: np.ndarray,
            y: np.ndarray
    ) -> float:
        if np.size(x) != np.size(y):
            raise Exception("'x' and 'y' must have the same size")
        if (
            np.any(x < 0)
            or np.any(y < 0)
            or np.any((x == 0) * (y != 0))
            or np.any((y == 0) * (x != 0))
        ):
            return np.inf
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
            y = np.reshape(y, (-1))
        mask = y > 0
        res = x[mask] * np.log(x[mask] / y[mask])
        return np.sum(res)

    def _check(self, x, gamma=1):
        if np.any(gamma <= 0):
            raise Exception("'gamma'  must be strictly positive")
        if np.size(gamma) > 1 and np.size(gamma) != np.size(x):
            ValueError(
                "'gamma' must be positive scalars or positive ND arrays" +
                " with the same size as 'x'"
            )



"""
Version : 1.0 (08-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Optional
import numpy as np


class Squashing:
    r"""Compute the proximity operator of the Squashing activation function.

    The squashing activation function is definded as:

           /  mu*arctan(sqrt(||x||/(mu - ||x||))-sqrt(||x|(mu-||x||) - x^2  if ||x||<mu
    f(x)= |  mu*( pi - mu )/2                                               if ||x||=mu
          \  +INF                                                           otherwise

        with mu = 8/(3*sqrt(3))

    Note: The proximity operator of this function f is the squashing function
    defined as:

            Squashing(y) = ( mu * ||y||/(1 +||y||) ) y

     INPUTS
    ========
     x     - ND array (np.ndarray)

     axis      - None or int, axis of block-wise processing [default: axis=None]
                  axis = None --> 'x' is processed as a single vector [DEFAULT] In this
                  case 'gamma' must be a scalar.
                  axis >=0   --> 'x' is processed block-wise along the specified axis
                  (0 -> rows, 1-> columns ect. In this case, 'gamma' must be singleton
                  along 'axis'.
    """

    def __init__(self, axis: Optional[int] = None):
        self.mu = 8 / (3 * np.sqrt(3))
        self.axis = axis

    def prox(self, x: np.ndarray) -> np.ndarray:
        mu = self.mu
        shape_sum = np.array(np.shape(x), dtype=int)
        shape_sum[self.axis] = 1
        l2_x = np.linalg.norm(x, ord=2, axis=self.axis).reshape(tuple(shape_sum))
        prox_x = (mu * l2_x / (1 + l2_x ** 2)) * x
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        mu = self.mu
        shape_sum = np.array(np.shape(x), dtype=int)
        shape_sum[self.axis] = 1
        l2_x = np.linalg.norm(x, ord=2, axis=self.axis).reshape(tuple(shape_sum))

        fun_x = mu * (np.pi - mu) / 2 * np.ones(np.shape(l2_x))
        fun_x[l2_x > mu] = np.inf
        mask = l2_x < mu
        fun_x[mask] = (
                mu * np.arctan(np.sqrt(l2_x[mask]) / (mu - l2_x[mask]))
                - np.sqrt(l2_x[mask] * (mu - l2_x[mask]))
                - 0.5 * l2_x[mask] ** 2
        )
        return np.sum(fun_x)



"""
Proximity operator of non-convex functions.

Created on Jun 2022

Author: Mbaye Diongue
"""

from .BurgCauchy import *
from .BurgLogSum import *
from .CapL1 import *
from .Cauchy import *
from .CEL0 import *
from .ConicL0 import *
from .EntropyCauchy import *
from .EntropyL0 import *
from .EntropyLogSum import *
from .L0L2Norm import *
from .L0Norm import *
from .LogSum import *
from .MCP import *
from .Root import *
from .SCAD import *
from .Truncated import *
from .TruncatedNorm import *

__all__ = [
    "BurgCauchy",
    "BurgLogSum",
    "CapL1",
    "Cauchy",
    "CEL0",
    "ConicL0",
    "EntropyCauchy",
    "EntropyL0",
    "EntropyLogSum",
    "L0L2Norm",
    "L0Norm",
    "LogSum",
    "MCP",
    "Root",
    "SCAD",
    "Truncated",
    "TruncatedNorm",
]



"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.newton import newton_


class Power:
    r"""Computes the proximity operator and the evaluation of gamma*f.

    Where f is the power function defined as:

              f(x) = x ^ q

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function, the output is a scalar (sum of the element-wise results)

    - But for the proximity operator (method 'prox'), the output has the same shape
    as the input 'x'.

     INPUTS
    ========
     x     - scalar or ND array
     q     -  scalar >= 1
     gamma - positive, scalar or ND array with the same size as 'x' (default: gamma=1)

    =======
    Examples
    ========

     Evaluate the function  f:

     >>> Power(3)( 2 )
     8

      Compute the result element-wise for vector inputs :

     >>> Power(q=2)( [-1, 3, 2] )
     14

     Compute the proximity operator at a given point :

     >>> Power(2).prox(  [-2, 3, 4 ])
     array([-0.66666667,  1.        ,  1.33333333])

     Use a scale factor 'gamma'>0 to compute the proximity operator of  the
     function 'gamma*f'

     >>> Power( 2, ).prox( [-2, 3, 4, 6 ], gamma=[1,2,3,3])
     array([-0.66666667,  0.6       ,  0.57142857,  0.85714286])
    """

    def __init__(self, q: float):
        if np.any(q < 1) or np.size(q) > 1:
            raise Exception("'q' must be a scalar greater than 1")
        self.q = q

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check(x, gamma)
        abs_x = np.abs(x)
        sign_x = np.sign(x)
        q = self.q

        if q == 1:
            prox_x = sign_x * np.maximum(0, abs_x - gamma)

        elif q == 4 / 3:
            ksi = np.sqrt(x**2 + 256 * gamma**3 / 729)
            ksi_minus = np.power(np.sqrt(ksi - x), 1 / 3)
            ksi_plus = np.power(np.sqrt(ksi + x), 1 / 3)

            prox_x = x + 4 * gamma / (3 * 2 ** (1 / 3)) * (ksi_minus - ksi_plus)

        elif q == 3 / 2:
            gamma2 = gamma**2
            prox_x = x + 9 / 8 * gamma * 2 * sign_x * (
                1 - np.sqrt(1 + 16 * abs_x / (9 * gamma2))
            )

        elif q == 2:
            prox_x = x / (2 * gamma + 1)

        elif q == 3:
            prox_x = sign_x * (np.sqrt(1 + 12 * gamma * abs_x) - 1)

        elif q == 4:
            ksi = np.sqrt(x**2 + 1 / (27 * gamma))
            prox_x = np.power((ksi + x) / (8 * gamma), 1 / 3) - np.power(
                (ksi - x) / (8 * gamma), 1 / 3
            )

        else:

            def polynom_phi(t):
                return gamma * q * t ** (q - 1) + t - abs_x

            def der_phi(t):
                return gamma * q * (q - 1) * t ** (q - 2) + 1

            # starting point
            prox_x = abs_x / 2
            prox_x = newton_(polynom_phi, der_phi, prox_x, 0, np.inf)
            prox_x = sign_x * prox_x

        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        return np.sum(np.abs(x) ** self.q)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it is an array) must "
                + "be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.shape(gamma) != np.shape(x)):
            raise Exception("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 (06-18-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class CapL1:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

    Where f is he capped-L1 (or truncated L1) penalty defined as:

               f(x) = min( 1, theta*|x| )

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    theta  - positive, scalar or ND array with the same size as 'x'[default: theta=1]

     Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>CapL1(theta=0.1)(x) will
    return a scalar even if x is a vector:

    >>>  CapL1(theta=0.1)(np.array([0, 2, -3]))
    0.5

    But as expected, >>>CapL1(theta=0.1).prox(x)
    will return a vector with the same size as x:

    >>> CapL1(theta=0.1).prox(np.array([0, 2, -3]))
    array([ 0. ,  1.9, -2.9])
    """

    def __init__(
            self,
            theta: Union[np.ndarray, float] = 1.0
    ):
        if np.any(theta <= 0):
            raise Exception("'theta'(or all of its elements) must be strictly positive")
        self.theta = theta

    def prox(self, x: np.ndarray, gamma: Union[np.ndarray, float] = 1.0) -> np.ndarray:
        self._check(x, gamma)
        theta = self.theta

        prox_x = (
                  (np.abs(x) < (1 / theta + gamma * theta / 2)) * (gamma * theta**2 < 2)
                  + np.sign(x) * (np.abs(x) - gamma * theta) *
                  (np.abs(x) > gamma * theta)
                  + x * (gamma * theta**2 < 2)
                  * (np.abs(x) >= (1 / theta + gamma * theta / 2))
        )
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        fun_x = np.minimum(1, self.theta * np.abs(x))
        return np.sum(fun_x)

    def _check(self, x, gamma):
        sz_x = np.size(x)
        if np.any(gamma <= 0):
            raise Exception("'gamma'(or all of its elements) must be strictly positive")
        if (np.size(self.theta) > 1) and (np.size(self.theta) != sz_x):
            raise Exception("'w' must be either scalar or the same size as 'x'")
        if (np.size(gamma) > 1) and (np.size(gamma) != sz_x):
            raise Exception("'gamma' must be either scalar or the same size as 'x'")



"""
Created on Wed Jun  8 2022

@author: Mbaye Diongue

Implementation of Newton to give the root of a funtion in a given interval [low, high]
"""

import numpy as np


def newton_(
        f,
        fp,
        x0: float or np.ndarray,
        low: float or np.ndarray = -1. * np.inf,
        high: float or np.ndarray = np.inf,
        max_iter: int = 100
) -> float or np.ndarray:
    """Find an annulation point of f in [low, high] using Newton's method.

    Parameters
    ----------
    f : a function (callable), we assume it has a root in [a, b].
    fp : derivative of f
    x0 : starting point
    low : lower bound of the interval in which we look the root
    high : upper bound of the interval in which we look the root
    max_iter: maximum number of iteration

    Returns
    -------
    x: root of the function f  i.e f(x)=0
    """
    tol = 1e-8
    val = 1e-20  # to avoid the division by zero when the derivative is null
    stop = False
    i = 0
    x = x0
    while (i < max_iter) and (not stop):
        x_old = x

        x = x - f(x) / (val + fp(x))  # newton step
        x = np.minimum(np.maximum(low, x), high)  # range constraint

        err = np.linalg.norm(x - x_old)
        if err < tol:
            stop = True
        i += 1

    return x



"""
Version : 1.0 (06-18-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class LogSum:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

      where the function f is defined as:


               f(x) = log( delta + |x|)


    'gamma' is the gamma factor

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    delta -positive, scalar or ND array with the same size as 'x'[default: delata=1e-5]
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]

    Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>LogSum(delta=0.01)(x) will
    return a scalar even if x is a vector:
    >>> LogSum(delta=0.01)(np.array([-1, 2, 3]))
    1.8100251316849367

    But as expected, >>>LogSum(delta=2).prox(x)
    will return a vector with the same size as x:
    >>> LogSum(delta=0.01).prox(np.array([-1, 2, 3]))
    array([0.        , 1.09512492, 2.6197333 ])
    """

    def __init__(
        self,
        delta: Union[float, np.ndarray] = 1e-5
    ):

        if np.any(delta <= 0):
            raise ValueError(
                "'delta' (or all of its components"
                + " if it is an array) must be strictly positive"
            )
        self.delta = delta

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components"
                + " if it is an array) must be strictly positive"
            )
        delta = self.delta
        # 1st branch
        prox_x = np.zeros_like(np.shape(x))

        # 2ND branch
        mask = np.abs(x) > np.sqrt(4 * gamma) - delta
        abs_x = np.abs(x[mask])
        if np.size(delta) > 1:
            delta = delta[mask]
        if np.size(gamma) > 1:
            gamma = gamma[mask]

        prox_x[mask] = (
            np.sign(x[mask])
            * 0.5
            * (abs_x - delta + np.sqrt((abs_x + delta) ** 2 - 4 * gamma))
        )
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        return np.sum(np.log(self.delta + np.abs(x)))



"""
Version : 1.0 (06-18-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class Truncated:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

      where the function f is defined as:

               f(x) = min{ x^2, w}

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    w     - positive, scalar or ND array with the same size as 'x'

    Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>Truncated(w=2)(x) will
    return a scalar even if x is a vector:

    >>> Truncated(w=2)(np.array([-1, 2, 3]))
    5

    But as expected, >>>Truncated(w=2).prox(x)
    will return a vector with the same size as x:

    >>> Truncated(w=2).prox(np.array([-1, 2, 3]))
    array([-0.33333333,  0.66666667,  3.        ])
    """

    def __init__(
            self,
            w: Union[float, np.ndarray]
    ):
        self.w = w

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        self._check_size(x, gamma)
        mask = x ** 2 < self.w * (1 + 2 * gamma)
        return x / (1 + 2 * gamma * mask)

    def __call__(self, x: np.ndarray) -> float:
        return np.sum(np.minimum(x ** 2, self.w))

    def _check_size(self, x, gamma):
        sz_x = np.size(x)
        if np.any(gamma <= 0):
            raise Exception(
                "'gamma' (or all of its components if it"
                + " is an array) must be strictly positive"
            )
        if (np.size(self.w) > 1) and (np.size(self.w) != sz_x):
            raise Exception("'a' must be either scalar or the same size as 'x'")
        if (np.size(gamma) > 1) and (np.size(gamma) != sz_x):
            raise Exception("'gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-22-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np
import numpy.fft as fft


class L2DownConv:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:

                        f(x) = || S_d*H*x - y||_2 ^2

    Where S_d is the down sampling operator with factor d>0 and H is a convolution
    operator that is described by the Fourier Kernel Lamb
    [i.e. Hx = ifftn( Lamb*fftn(x) )]

     INPUTS
    ========
     x      - ND array
     Lamb   - ND array with the same size as 'x' ( Fourier kernel of the
                convolution operator H)
     d      - int or vector of positive integers (down sampling in each dimension of x)
     gamma  - positive scalar
     y      - ND array with size : size(x)/d
    """

    def __init__(
            self,
            lamb: np.ndarray,
            d: int or np.ndarray,
            y: np.ndarray
    ):
        d = np.array(d, dtype=int)
        self.lamb = lamb
        self.d = d
        self.y = y

        # pre-Computations
        sz = np.shape(lamb)
        mask_ = np.zeros(np.size(lamb), dtype=bool)

        if np.size(d) <= 1:
            mask_[np.arange(1, sz[0], d)] = True
        else:
            for ii in range(np.size(d)):
                sample = np.zeros(sz[ii], dtype=bool)
                sample[np.arange(0, sz[ii], d[ii])] = True
                mask_[ii * sz[ii]:(ii + 1) * sz[ii]] = sample
        self.mask_ = mask_.reshape(np.shape(lamb))

    def prox(self, x: np.ndarray, gamma: float = 1.0) -> np.ndarray:
        self._check(x, gamma)
        y = self.y
        d = self.d
        lamb = self.lamb
        y_up = np.zeros(np.shape(x))
        y_up[self.mask_] = np.reshape(y, (-1))[0:np.sum(1 * self.mask_)]
        sz_ = np.array(np.shape(x))
        sz_patch = np.array(sz_ / d, dtype=int)

        def expand_mat(mat: np.ndarray, size_patch: np.ndarray) -> np.ndarray:
            res: np.ndarray = np.zeros(size_patch)
            sz_2 = []
            for k in range(len(size_patch)):
                sz_2.append(size_patch[k])
            sz_2.append(-1)
            sz_2 = tuple(sz_2)
            tmp = np.reshape(mat, sz_2)
            for n in range(np.shape(tmp)[-1]):
                res = res + tmp[..., n]
            return res

        h2_h2t = expand_mat(np.abs(lamb) ** 2, sz_patch)

        def rep_mat(xx: np.ndarray):
            return np.tile(xx, d)

        # Prox computation
        fft_reverse = np.conj(lamb) * np.fft.fftn(y_up) + np.fft.fftn(x / gamma)

        prox_x = fft.ifftn(
            gamma
            * (
                    fft_reverse
                    - np.conj(lamb)
                    * rep_mat(expand_mat(lamb * fft_reverse,
                                         sz_patch) / (np.size(d) / gamma +
                                                      h2_h2t)).reshape(np.shape(x))
            )
        )

        return prox_x.real

    def __call__(self, x: np.ndarray) -> float:
        hx = fft.fftn(self.lamb * np.fft.ifftn(x))
        hx = hx.real  # the real part of the Fourier transform
        yy = np.reshape(self.y, (-1))[0:np.sum(1 * self.mask_)]
        res = 0.5 * np.sum(hx[self.mask_] - yy) ** 2
        return res

    def _check(self, x, gamma=1):
        if np.size(gamma) > 1 or gamma <= 0:
            raise Exception("'gamma' must be a strictly positive scalar")
        if np.any(np.array(np.shape(self.y)) != np.array(np.shape(x)) / self.d):
            raise Exception("'y' must be an ND array with shape: np.shape(x)/d")
        if np.size(self.lamb) != np.size(x):
            raise Exception("'Lamb' must be an ND array with size as 'x'")



"""
Version : 1.0 (06-18-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class MCP:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

    Where f is the Minimax Concave Penalty (MCP) defined as:


                  |  lamb^2*beta/2                    if |x| > beta*lamb
        f(x) =   |
                 |  lamb*|x| - x^2/(2*beta)            if |x| <= a*lamb

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    lamb  - positive, scalar or ND array with the same size as 'x'
    beta  - positive, scalar or ND array with the same size as 'x'
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]


     Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>MCP(lamb=2)(x) will
    return a scalar even if x is a vector:

    >>> MCP(lamb=2)(np.array([-1, 2, 3]))
    5.5

    But as expected, >>>MCP(lamb=2).prox(x)
    will return a vector with the same size as x:

    >>> MCP(lamb=2).prox(np.array([-1, 2, 3]))
    array([0, 0, 3]
    """

    def __init__(
        self,
        lamb: Union[float, np.ndarray],
        beta: Union[float, np.ndarray]
    ):

        if np.any(lamb <= 0):
            raise ValueError(
                "'lamb' (or all of its components if it "
                + "is an array) must be strictly positive"
            )

        if np.any(beta <= 0):
            raise ValueError(
                "'beta' (or all of its components if it"
                + " is an array) must be strictly positive"
            )
        self.lamb = lamb
        self.beta = beta

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        self._check_size(x, gamma)
        beta = self.beta
        lamb = self.lamb
        abs_x = np.abs(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        mask1 = np.abs(x) > np.sqrt(gamma * beta) * lamb
        prox_x = x * mask1
        if np.size(beta - gamma) <= 1 and (beta <= gamma):
            return np.reshape(prox_x, np.shape(x))

        mask = beta > gamma
        if np.size(lamb) > 1:
            lamb = lamb[mask]
        if np.size(beta) > 1:
            beta = beta[mask]
        if np.size(gamma) > 1:
            gamma = gamma[mask]
        filter1 = beta / (beta - gamma) * np.maximum(abs_x[mask] - lamb * gamma, 0)
        prox_x[mask] = np.sign(x[mask]) * np.minimum(filter1, abs_x[mask])

        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        self._check_size(x, 1)
        beta = self.beta
        lamb = self.lamb
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        # 1st branch
        abs_x = np.abs(1.0 * x)
        fun_x = lamb * abs_x - 0.5 * x**2 / beta

        # 2nd branch
        mask = abs_x > beta * lamb
        if np.size(lamb) > 1:
            lamb = lamb[mask]
        if np.size(beta) > 1:
            beta = beta[mask]
        fun_x[mask] = 0.5 * beta * lamb**2

        return np.sum(fun_x)

    def _check_size(self, x, gamma):
        sz_x = np.size(x)
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it"
                + " is an array) must be strictly positive"
            )
        if (np.size(self.lamb) > 1) and (np.size(self.lamb) != sz_x):
            raise ValueError("'lamb' must be either scalar or the same size as 'x'")
        if (np.size(self.beta) > 1) and (np.size(self.beta) != sz_x):
            raise ValueError("'beta' must be either scalar or the same size as 'x'")
        if (np.size(gamma) > 1) and (np.size(gamma) != sz_x):
            raise ValueError("'gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-22-2022).

DEPENDENCIES:
    - 'Max.py' located in the folder 'multi'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Optional
import numpy as np
from proxop.multi.Max import Max


class Linf(Max):
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the infinitive norm:

                        f(x) = gamma *tau* max( |x1|, ..., |xn|)

    INPUTS
    ========
     x         - ND array
     gamma     - positive, scalar or ND array compatible with the blocks of 'x'
                 [default: gamma=1], 'gamma' is the scale factor
     axis    - None or int, axis of block-wise processing [default: axis=None]
                  axis = None --> 'x' is processed as a single vector [DEFAULT] In this
                  case 'gamma' must be a scalar.
                  axis >=0   --> 'x' is processed block-wise along the specified axis
                  (0 -> rows, 1-> columns ect. In this case, 'gamma' must be singleton
                  along 'axis'.
    """

    def __init__(
            self,
            axis: Optional[int] = None
    ):
        super().__init__(axis=axis)

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        abs_x = np.abs(x)
        return np.sign(x) * Max(axis=self.axis).prox(abs_x, gamma)

    def __call__(self, x: np.ndarray) -> float:
        linf_x = np.max(np.abs(x), self.axis)
        return np.sum(linf_x)



"""
Version : 1.0 ( 06-20-2022).

DEPENDENCIES:
     - 'NuclearNorm.py' located in the folder 'multi'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.multi.NuclearNorm import NuclearNorm


class NuclearBlocks:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:


                f(x) = gamma * \sum_{i= 1}^N \sum_{j=1}^M w_{i,j} \|X_{i,j}\|_N

            where X = U*diag(s)*V.T \in R^{M*N}

     INPUTS
    ========
     x        -  (M,N) -array_like ( representing an M*N matrix )

     ind_r    - Vector partitioning the rows of X in groups EXAMPLE: ind_r=[1 2 2 3 3 1]
     means there are max(ind_r)=3 blocks:  the first block contains the first and last
     rows of x, the 2nd block contains the 2nd and the third row, the third block
     contains the 4th and the 5th row.

     ind_c    - Vector partitioning the columns of X in groups
     W        - positive ND array of size (max(ind_r) , max(ind_c) )
     gamma    - positive, scalar [default: gamma=1]
    """

    def __init__(
            self,
            w: Union[float, np.ndarray],
            ind_r: np.ndarray,
            ind_c: np.ndarray
    ):
        if np.any(w <= 0) or np.any((np.max(ind_r), np.max(ind_c)) != np.shape(w)):
            raise Exception(
                "'W' must be positive and of shape '( max(ind_r), max(ind_c))'"
            )
        self.w = w
        self.ind_r = np.array(ind_r)
        self.ind_c = np.array(ind_c)

    def prox(
            self,
            x: np.ndarray,
            gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        self._check(x, gamma)
        ind_r = self.ind_r
        ind_c = self.ind_c
        w = self.w

        m, n = np.shape(x)
        if np.size(ind_r) != m or np.size(ind_c) != n:
            raise ValueError("'ind_r' or 'ind_c' has not the right length: " +
                             "It is expected: (len(ind_r), len(ind_c)) = np.shape(x)")
        prox_x = np.zeros((m, n))
        for i in range(max(ind_r)):
            for j in range(max(ind_c)):
                ind_i = ind_r == i
                ind_j = ind_c == j
                mask = np.dot(ind_i.T, ind_j)
                # The elements that don't belong to the block are replaced by zeros
                x_mask = x * mask

                # Update of p
                prox_x = prox_x + NuclearNorm().prox(x_mask, w[i, j] * gamma)

        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        ind_r = self.ind_r
        ind_c = self.ind_c
        fun_x = 0
        for i in range(max(ind_r)):
            for j in range(max(ind_c)):
                ind_i = ind_r == i
                ind_j = ind_c == j
                mask = np.dot(ind_i.T, ind_j)
                # The elements that don't belong to the block are replaced by zeros
                x_mask = x * mask
                # Update of fun_x
                fun_x = fun_x + NuclearNorm()(x_mask)
        return fun_x

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise Exception(
                "'gamma' (or all of its components if it is an array)" +
                " must be strictly positive"
            )
        if len(np.shape(x)) != 2:
            raise ValueError(
                "'x' must be an (M,N) -array_like ( representing a M*N matrix )"
            )



"""
Version : 1.0 ( 06-23-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class BregExp:
    r"""Compute the proximity operator prox_{gamma*f}^{phi}.

    where the functions f and phi are defined as follows:

              f(x)=  exp(x)

              phi(u)= exp(u)

    Note: The function phi is a Legendre type function and induces a Bregman distance

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    - When calling the functions f or phi (method 'phi') the result is computed
    element-wise SUM.
    So the command >>>BregExp()(x) or >>>BregExp().phi(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'.So, the command >>>BregExp.prox(x)   will return
    an array with the same shape as 'x'.

    INPUTS
    ========
    x     - scalar or ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]

    =======
    Examples
    ========

    Evaluate the function f:

    >>> BregExp()( np.log(2) )
    2.0
    >>> BregExp()( [-2, 3, 4, np.log(3)] )
    77.81902223956851

     Evaluate The function  phi

     >>> BregExp().phi( [np.log(3)] )
     3.0000000000000004

     Compute the proximity operator at a given point :

     >>> BregExp().prox(  [-2, 3, 4, np.log(3)] )
      array([-2.69314718,  2.30685282,  3.30685282,  0.40546511])
     >>> BregExp().prox(  [-2, 3, 4, np.log(3)] , gamma=2)
     array([-3.09861229,  1.90138771,  2.90138771,  0.        ])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        return x - np.log(1 + gamma)

    def __call__(self, x: np.ndarray) -> float:
        return np.sum(np.exp(x))

    def phi(self, u: np.ndarray) -> float:
        return np.sum(np.exp(u))

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError("'gamma'  must be strictly positive")
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-21-2022).

DEPENDENCIES:
     -'prox_svd.py' - located in the folder 'utils'
     -'fun_svd.py'  - located in the folder 'utils'
     -'Linf.py'     - located in the folder 'multi'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.multi.Linf import Linf
from proxop.utils.prox_svd import prox_svd
from proxop.utils.fun_svd import fun_svd


class Spectral:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:

                         f(x)= ||X||_S

            where X = U*diag(s)*V.T \in R^{M*N}


     INPUTS
    ========
     x         -  (M,N) -array_like ( representing an M*N matrix )
     gamma     - positive, scalar or ND array compatible with the size of 'x'
                 [default: gamma=1]
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        if np.any(gamma <= 0) or np.size(gamma) > 1:
            raise Exception("'gamma'  must be a strictly positive scalar")
        return prox_svd(x, gamma, Linf().prox)

    def __call__(self, x: np.ndarray) -> float:
        return fun_svd(x, 1, Linf())



"""
Version : 1.0 ( 06-23-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class BregAbsEntropy:
    r"""Compute the proximity operator prox_{gamma*f}^{phi}.

    Where the functions f and phi are defined as follows:

              f(x)=  | x - delta |

                      /  u*log(u)     if u>0
             phi(u)= |   0            if u=0
                     \   +inf         otherwise

    Note: The function phi is a Legendre type function and induces a Bregman distance

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    - When calling the functions f or phi (method 'phi') the result is computed
    element-wise SUM.
    So the command >>>BregAbsEntropy()(x) or >>>BregAbsEntropy().phi(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the
    same shape as the input 'x'.
    So, the command >>>BregAbsEntropy.prox(x)   will return an array with the same
    shape as 'x'

    INPUTS
    ========
    x          - scalar or ND array
    delta      - scalar or ND array with the same size as 'x'
    gamma      - scalar or ND array compatible with the blocks of 'y' [default: gamma=1]

    ========
     Examples
     ========

     Evaluate the  function f :

     >>> BregAbsEntropy(delta=0.001)(4)
     3.999
     >>> BregAbsEntropy(delta=0.001)( [-2, 3, 4] )
     8.999

     Evaluate The function  phi

     >> BregAbsEntropy(delta=0.001).phi([2, 3, 4])
     10.227308671

     Compute the proximity operator at a given point :

     >>> BregAbsEntropy(delta=0.1).prox( [-1, 2, 4] )
     array([-2.71828183,  0.73575888,  1.47151776])
     >>> BregAbsEntropy(delta=0.1).prox( [-1, 2, 4], gamma=2 )
     array([-7.3890561 ,  0.27067057,  0.54134113]
    """

    def __init__(self, delta: Union[float, np.ndarray]):
        if np.size(delta) > 1 and (not isinstance(delta, np.ndarray)):
            delta = np.array(delta)
        if np.any(delta <= 0):
            raise ValueError("'delta'  must be strictly positive")
        self.delta = delta

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        delta = self.delta
        self._check(x, gamma)

        # scalar-like inputs handling
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        # 3rd branch
        prox_x = x * np.exp(-gamma)

        # 2nd branch
        mask1 = x >= delta * np.exp(-gamma)
        mask = mask1 * (x <= delta * np.exp(gamma))
        dd = delta
        if np.size(dd) > 1:
            dd = dd[mask]
        prox_x[mask] = dd

        # 1st branch
        mask = np.logical_not(mask1)
        gg = gamma
        if np.size(gg) > 1:
            gg = gg[mask]
        prox_x[mask] = x[mask] * np.exp(gg)

        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        return np.sum(np.abs(x - self.delta))

    def phi(self, u: np.ndarray) -> float:
        if not isinstance(u, np.ndarray):
            u = np.array(u)
        if np.any(u < 0):
            return np.inf
        if np.size(u) <= 1:
            u = np.reshape(u, (-1))
        res = np.zeros(np.shape(u))
        mask = u > 0
        res[mask] = u[mask] * np.log(u[mask])

        return np.sum(res)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError("'gamma'  must be strictly positive")
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or have " +
                             "the same size as 'x'")



"""
Version : 1.0 ( 06-16-2022).

DEPENDENCIES:
     - 'newton.py' located in the folder 'utils'

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np
from proxop.utils.newton import newton_


class EpiSquaredHingeLoss:
    r"""Compute the projection and the indicator of the epigraph of phi.

    Where phi is the function defined as:
                phi(y)= w * max( y, 0) ^ 2

    When the inputs are arrays, the outputs are computed element-wise
     INPUTS
    ========
     y    - ND array
     kxi  - ND array with the same size as 'y'
     w    - positive, scalar or ND array with the same size as 'ksi'
    """

    def __init__(
            self,
            ksi: np.ndarray,
            w: Union[float, np.ndarray] = 1.0
    ):
        if np.any(w <= 0):
            raise ValueError(
                "'w' (or all of its components if it is an array) must be positive"
            )
        if np.size(w) > 1 and (np.size(w) != np.size(ksi)):
            raise ValueError(" 'w' must be a scalar or must have" +
                             " the same size as 'ksi'")
        if np.size(ksi) <= 1:
            ksi = np.reshape(ksi, (-1))
        self.w = w
        self.ksi = ksi

    def prox(self, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        self._check(y)
        ksi = self.ksi
        w = self.w

        if np.size(y) <= 1:
            y = np.reshape(y, (-1))

        # Note: same computation as in 'EpiPower.py' for q=2
        def fun_phi(t):
            return 2 * w**2 * t**3 + (1 - 2 * w * ksi) * t - np.abs(y)

        def der_phi(t):
            return 6 * w**2 * t**2 - 2 * w * ksi + 1

        # starting point
        low = (np.maximum(ksi, 0) / w) ** (1 / 2)
        root_init = 3 * low

        # Finding the root of the polynom with the Newton method
        root = newton_(fun_phi, fp=der_phi, x0=root_init, low=low, high=np.inf)
        prox_t = w * root ** 2
        prox_p = np.sign(y) * root

        # 1st and 2nd branches
        mask = np.logical_or(w * np.maximum(y, 0) <= ksi, y <= 0)
        prox_p[mask] = y[mask]
        prox_t[mask] = np.maximum(ksi[mask], 0)
        return tuple([prox_p, prox_t])

    def __call__(self, y: np.ndarray) -> float:
        """Indicate if the input 'x' is in the constraint set or not.

        Parameters
        ----------
        x : np.ndarray

        Returns
        -------
        0      if the input 'x' is in the set
        +inf   otherwise
        """
        if np.all(self.w * np.maximum(y, 0) ** 2 <= self.ksi):
            return 0
        return np.inf

    def _check(self, y):
        if np.size(y) != np.size(self.ksi):
            raise ValueError("'y' must have the same size as 'ksi'")



from .scalar import *
from .nonconvex import *
from .multi import *
from .indicator import *
from . import scalar
from . import multi
from . import nonconvex
from . import indicator

__version__ = "1.0.3"



"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np


class LogActi:
    r"""Compute the proximity operator and the evaluation of f.

    Where f is the logarithmic activation function defined as:

            f(x) = cosh(x)-(x^2)/2

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>LogActi()(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'.
    So, the command >>>LogActi().prox(x)   will return an array with the same
    shape as 'x'

     INPUTS
    ========
     x     - scalar or ND array

     =======
     Examples
     ========

     >>> LogActi()( -3 )
      11.585536923187668
     >>> LogActi()( [1, -1.5, -0.5] )
      1.0986921694972378

     Compute the proximity operator at a given point :

     >>> LogActi().prox(-3)
     -1.3862943611198906
     >>> LogActi().prox(  [-2, 3, -0.5] )
     array([-1.09861229,  1.38629436, -0.40546511])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray) -> np.ndarray:
        return np.sign(x) * np.log(1 + np.abs(x))

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        return np.sum(np.exp(np.abs(x)) - np.abs(x) - 1 - 0.5 * x**2)



"""
Version : 1.0 ( 06-23-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class BregSquareLog:
    r"""Compute the proximity operator prox_{gamma*f}^{phi}.

    Where f and phi are defined as follows:

               f(x)= (1/2)*x^2

                         /  - log(u)      if u>0
              phi(u) =  |
                        \   +inf           otherwise

    Note: The function phi is a Legendre type function and induces a Bregman distance

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    - When calling the functions f or phi (method 'phi') the result is computed
    element-wise SUM. So the command >>>BregSquareLog()(x) or >>>BregSquareLog.phi(x)
    will return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'.
    So, the command >>>BregSquareLog.prox(x)   will return an array with the same
    shape as 'x'

    INPUTS
    ========
    x      - scalar or ND array
    gamma  - positive, scalar or ND array with the same size as 'x' [default: gamma=1]


    =======
    Examples
    ========

    Evaluate the function f:

     >>> BregSquareLog()( -3 )
     4.5
     >>> BregSquareLog()( [1, 2, 3] )
     7.0

     Evaluate The function  phi

     >> BregSquareLog().phi( [2, 3, 4] )
     -3.1780538303479453

     Compute the proximity operator at a given point :

     >>> BregSquareLog().prox(  [-2, 3, 4, np.e] )
     array([-0.78077641,  0.84712709,  0.88278222,  0.83283647])
     >>> BregSquareLog().prox(  [-2, 3, 4, np.log(2)] )
     array([-0.78077641,  0.84712709,  0.88278222,  0.51167407])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check(x, gamma)
        x[x == 0] = 1e-20
        return (np.sqrt(1 + 4 * gamma * x**2) - 1) / (2 * gamma * x)

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        return np.sum(0.5 * x**2)

    def phi(self, u: np.ndarray) -> float:
        if np.size(u) > 1 and (not isinstance(u, np.ndarray)):
            u = np.array(u)
        if np.size(u) <= 1:
            u = np.reshape(u, (-1))
        res = np.zeros(np.shape(u))
        mask = u > 0
        res[mask] = -np.log(u[mask])
        res[u <= 0] = np.inf
        return np.sum(res)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError("'gamma'  must be strictly positive")
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise Exception("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class LogisticLoss:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the logistic loss function defined as:

         f(x)= log( 1+ exp(x) )

    'gamma' is the scale factor

     When the input 'x' is an array, the output is computed element-wise :

     -When calling the function, the output is a scalar (sum of the element-wise
    results ).

     - But for the proximity operator (method 'prox'), the output has the same shape
     as the input 'x'.

      INPUTS
     ========
      x     - ND array
      gamma - positive, scalar or ND array with the same size as 'x' (default: gamma=1)

     =======
     Examples
     ========

      Evaluate the function  f:

      >>> LogisticLoss()( 0 )
       0.6931471805599453

       Compute the result element-wise sum for vector inputs :

      >>> LogisticLoss()( [-1, 3, np.e])
      18.432098909466088

      Compute the proximity operator at a given point :

      >>> LogisticLoss().prox(  [-2, 3, 4 ])
      array([-2.10829336,  2.10829336,  3.0454158 ]))

      Use a scale factor 'gamma'>0 to compute the proximity operator of
      the function 'gamma*f'

      >>> LogisticLoss().prox( [-2, 3, 4, np.e ], gamma=2.5)
      array([-1.04044264, -0.68339574, -0.55797119, -0.70389204])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check(x, gamma)
        scale = gamma

        x = (
            x / scale
        )  # we do this scaling to use the same method as in FermiDiracEntropy.py class
        gamma = 1 / scale

        limit = 8
        # RUN HALEY'S METHOD TO SOLVE w = W_{exp(x/gamma)}(exp(x/gamma)/gamma)
        # where W_r(x) is the generalized Lambert function

        w = np.zeros_like(1.0 * x)
        inv_gamma = 1 / gamma
        c = x * inv_gamma - np.log(gamma)
        z = np.exp(c)
        r = np.exp(x * inv_gamma)

        # ASYMPTOTIC APPROX
        approx = 1 - np.exp((1 - x) * inv_gamma)

        # INITIALIZATION
        # Case 1: gamma <= 1/30
        mask = np.logical_and(z > 1, gamma <= 1 / 30)
        w[mask] = c[mask] - np.log(c[mask])
        # Case 2: gamma > 1/30
        mask = np.logical_and(z > 1, gamma > 1 / 30)
        if np.size(inv_gamma) > 1:
            inv_gamma = inv_gamma[mask]
        w[mask] = inv_gamma * approx[mask]

        # RUN
        max_iter = 20
        test_end = np.zeros(np.shape(x))
        tol = 1e-8
        epsilon = 1e-20  # to avoid dividing by zero
        n = np.size(x)

        for _ in range(max_iter):

            e = np.exp(w)
            y = w * e + r * w - z
            v = e * (1 + w) + r
            u = e * (2 + w)
            w_new = w - y / (v - y * u / (2 * v))

            mask = np.abs(w_new - w) / (epsilon + np.abs(w)) < tol
            test_end[np.logical_and(mask, test_end == 0)] = 1
            not_mask = np.logical_not(mask)
            idx_update = np.logical_and(not_mask, test_end == 0)
            w[idx_update] = w_new[idx_update]  # the rest stays constant !
            if np.sum(test_end) == n:  # stop !
                break

        prox_x = gamma * w

        if np.size(prox_x) <= 1:
            if c > limit and gamma > 1:
                prox_x = approx
            return x - scale * np.minimum(prox_x, 1)

        # ASYMPTOTIC DVP
        mask_approx = np.logical_and(c > limit, gamma > 1)
        prox_x[mask_approx] = approx[mask_approx]

        # FINAL THRESHOLD TO AVOID NUMERICAL ISSUES FOR SMALL GAMMA
        prox_x = np.minimum(prox_x, 1)

        return x - scale * prox_x

    def __call__(self, x: np.ndarray) -> float:
        return np.sum(np.log(1 + np.exp(x)))

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it is an array) must "
                + "be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-21-2022).

DEPENDENCIES:
     -'prox_svd.py' - located in the folder 'utils'
     -'fun_svd.py'  - located in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np
from proxop.utils.prox_svd import prox_svd
from proxop.utils.fun_svd import fun_svd


class SquaredFrobeniusNormLogDet:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the funtion definded as:

               / -log( det(X) ) + mu*||X||_F^2    if X is a symmetric positive
         f(x)=|                                   definite matrix
              \  + inf                            otherwise

    where

           * X = U*diag(s)*V.T \in R^{M*N}  the Singular Value decomposition of X

           * det(X)= prod_i(s_i)   is the determinant of the matrix X

           * ||X||_F^2 = ||s||_2^2  the squared of the Frobenius norm of X

          * 'gamma*tau' is the scale factor

     Note:
        No checking is performed to verify whether X is symmetric or not when computing
        the proximity operator with the method 'prox'. X is assumed to
        be symmetric.

     INPUTS
    ========
     x               -  (M,N) -array_like ( representing an M*N matrix)
     mu              - positive scalar  [default: mu=1]
     gamma           - positive scalar  [default: gamma=1]
    """

    def __init__(self, mu: float = 1):
        if np.any(mu < 0) or np.size(mu) > 1:
            raise Exception("'mu'  must be a positive scalar")
        self.mu = mu

    def prox(self, x: np.ndarray, gamma: float = 1) -> np.ndarray:
        self._check(x, gamma)
        mu = self.mu

        def prox_phi(s, gam):
            return (s + np.sqrt(s**2 + 4 * gam * (2 * mu * gam + 1))) / (
                2 * (2 * gam * mu + 1)
            )

        return prox_svd(x, gamma, prox_phi, hermitian=True)

    def __call__(self, x: np.ndarray) -> float:
        tol = 1e-10
        # Check if the matrix is symmetric
        if not np.allclose(x, np.transpose(x)):
            return np.inf

        def fun_phi(s):
            if np.any(s <= tol):
                return np.inf
            return -np.log(np.prod(s)) + self.mu * np.sum(s**2)

        return fun_svd(x, 1, fun_phi)

    def _check(self, x, gamma):
        if np.any(gamma <= 0) or np.size(gamma) > 1:
            raise Exception("'gamma'  must be a strictly positive scalar")
        if len(np.shape(x)) != 2:
            raise ValueError(
                "'x' must be an (M,N) -array_like ( representing an M*N matrix )"
            )



"""
Version : 1.0 ( 06-16-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Optional, Tuple
import numpy as np


class EpiMax:
    r"""Compute the projection and the indicator of the epigraph of phi.

    Where phi is the function defined as:

                phi(y)= w * max( x1, ..., x_n)

     INPUTS
    ========
     y    - ND array
     ksi  - ND array compatible with the blocks of 'y' (see above)
     w    - positive, scalar or ND array with the same size as 'ksi' [default: w=1]
     axis - int or None, direction of block-wise processing
            When the input 'y' is an array, the computation can vary as follows:
            - axis = None --> 'y' is processed as a single vector [DEFAULT]
                                   (in this case, 'xi' must be scalar)
            - axis >= 0 --> 'y' is processed block-wise along the specified axis
            (in this case, 'xi' must be singleton along 'axis')
    """

    def __init__(
            self,
            ksi: np.ndarray,
            axis: Optional[int] = None,
            w: Union[float, np.ndarray] = 1.0
    ):

        if np.any(w <= 0):
            raise ValueError("'w' must be positive")
        if np.size(w) > 1 and (np.size(w) != np.size(ksi)):
            raise ValueError(" 'w' must be a scalar or have the same size as 'ksi'")
        if np.size(ksi) <= 1:
            ksi = np.array([ksi])

        self.w = w
        self.axis = axis
        self.ksi = ksi

    # proximal operator (i.e. the projection on the constraint set)
    def prox(self, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        self._check(y)
        ksi = self.ksi
        w = self.w
        axis = self.axis
        if len(np.shape(y)) == 1 and np.size(ksi) <= 1:
            axis = None

        if axis is not None:
            if axis == 0:
                ksi = np.reshape(ksi, (2 * axis + 1) * (1, -1))
            else:
                ksi = np.reshape(ksi, (-1, 1))
        # sort inputs
        v = y * w
        v = np.sort(v, axis=axis)

        # compute the cumulative sum
        cum_sum = np.concatenate((w**2 * ksi, np.flip(v, axis=axis)), axis=axis)
        cum_sum = np.cumsum(cum_sum, axis=axis)
        cum_sum = np.flip(cum_sum, axis=axis)

        axis_ = axis
        if axis is None:
            axis_ = 0
        # Normalize...
        cum_sum = cum_sum / (
            w**2 + np.shape(v)[axis_] + 1 - np.cumsum(np.ones_like(cum_sum), axis=axis)
        )

        infty = np.inf * np.ones(np.shape(ksi))
        v_low = np.concatenate((-infty, v), axis=axis)
        v_high = np.concatenate((v, infty), axis=axis)

        mask = np.logical_and(v_low < cum_sum, cum_sum <= v_high)

        sz_ = np.shape(mask)[axis_]
        mat = (np.arange(np.size(mask)) % sz_).reshape(np.shape(mask))
        mask = mask * mat

        ind_max = np.max(mask, axis=axis)

        if axis is not None:
            prox_t = np.reshape(cum_sum[ind_max], np.shape(ksi))
        else:
            prox_t = (np.reshape(np.diag(np.take(cum_sum, ind_max, axis=axis)),
                                 np.shape(ksi)))

        prox_y = np.minimum(y, prox_t / w)

        return tuple([prox_y, prox_t])

    # indicator of the constraint set
    def __call__(self, y) -> float:
        self._check(y)
        tol = 1e-10
        if self.axis is None:
            y = np.reshape(y, (-1))
        mask = self.w * np.max(y, axis=self.axis) <= self.ksi + tol
        if np.all(mask):
            return 0
        return np.inf

    def _check(self, y):
        sz_ = 1
        if self.axis is not None:
            sz = np.array(np.shape(y), dtype=int)
            sz_ = np.delete(sz, self.axis)
        cond1 = self.axis is not None and np.any(sz_ != np.shape(self.ksi))
        cond2 = self.axis is None and np.any(np.size(self.ksi) != sz_)
        if cond1 or cond2:
            try:
                self.ksi = np.reshape(self.ksi, sz_)
            except ValueError:
                print("'ksi' is not compatible with the size of 'y' ")



"""
Version : 1.0 (06-18-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class L0Norm:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

      where the function f is defined as:

                                 / 0   if x=0
               f(x) = ||x||_0 = |
                                \  1   otherwise

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tol   - float, tolerance of the L0 norm ( if |x_i|<tol x_i is assumed to be null)
            [default: tol=1e-10]

    Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>L0Norm(gamma=1)(x) will return
    a scalar even if x is a vector:
    >>> L0Norm(gamma=1)(np.array([-1.,2.,3]))
    3

    But as expected, >>>L0Norm().prox(x) will
    return a vector with the same size as x:

    >>> L0Norm().prox(np.array([-1.,2.,3]))
    array([-0.,  2.,  3.])

    """

    def __init__(
            self,
            tol: float = 1e-10
    ):
        self.tol = tol

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        self._check(x, gamma)
        mask = x**2 >= 2 * gamma
        return x * mask

    def __call__(self, x: np.ndarray) -> float:
        return np.sum(np.abs(x) >= self.tol)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise Exception(
                "'gamma' (or all of its elements"
                + " if it is an array) must be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise Exception("'gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-14-2022).

DEPENDENCIES:
     - 'pava_python.py' in the folder 'utils'

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np
from proxop.utils.pava_python import pava_python


class MonotoneCone:
    """Compute the projection and the indicator of the monotone cone.

    The 'monotone cone' is the set containing all vectors X that verify:

                               x1 <= x2 <= ... <= xN


     where (1, ..., 1) is a ND array with all components equal to one,
     and (1,..., 1).T denotes its transpose

     INPUTS
    ========
     x    - ND array
    """

    # proximal operator (i.e. projection on the monotone cone)
    def __init__(self):
        pass

    def prox(self, x: np.ndarray) -> np.ndarray:
        return pava_python(x)

    # indicator of the monotone cone
    def __call__(self, x: np.ndarray) -> float:
        """
        Indicate if the input 'x' is in the constraint set or not
        Parameters
        ----------
        x : np.ndarray

        Returns
        -------
        0      if the input 'x' is in the set
        +inf   otherwise
        """
        if np.size(x) <= 1:
            return 0
        if np.all(x[:-1] <= x[1:]):
            return 0
        return np.inf



"""
Version : 1.0 (06-18-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class SCAD:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

    where f is Smoothly Clipped Absolute Deviation (SCAD)  defined as:


                  |  lamb*|x|                               if |x| <= lamb
        f(x) =   |  -(lamb^2-2*a*lamb*|x| + x^2)/[2*(a-1)]  if lamb < |x| <= a*lamb
                 |  (a+1)*lamb^2/2                          if |x| > a*lamb

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    lamb  - positive, scalar or ND array with the same size as 'x'
    a     - >2, scalar or ND array with the same size as 'x'
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]

     Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>SCAD(lamb=2, a=3)(x) will
    return a scalar even if x is a vector:

    >>> SCAD(lamb=2, a=3)(np.array([-1, 2, 3]))
    11.75

    But as expected, >>>SCAD(lamb=2, a=3).prox(x)
    will return a vector with the same size as x:

    >>> SCAD(lamb=2, a=3).prox(np.array([-1, 2, 3]))
    array([0., 0., 1.])
    """

    def __init__(
        self,
        lamb: Union[float, np.ndarray],
        a: Union[float, np.ndarray]
    ):

        if np.any(lamb <= 0):
            raise Exception(
                "'lamb' (or all of its components if it"
                + " is an array) must be strictly positive"
            )
        if np.any(a <= 2):
            raise Exception(
                "'a' (or all of its components if it"
                + " is an array) must be greater than 2"
            )
        self.lamb = lamb
        self.a = a

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        self._check_size(x, gamma)
        a = self.a
        lamb = self.lamb
        abs_x = np.abs(x)
        sign_x = np.sign(x)

        # 1st case:  a <= 1+gamma
        mask1 = abs_x <= 0.5 * (a + 1 + gamma) * lamb
        mask2 = 1 - mask1
        prox_x = 1.0 * sign_x * np.maximum(abs_x - lamb * gamma, 0) * mask1 + x * mask2

        if np.size(a - gamma) <= 1 and (a <= 1 + gamma):
            return np.reshape(prox_x, np.shape(x))

        # 2nd case:  a>1+gamma
        mask = a > 1 + gamma
        if np.size(lamb) > 1:
            lamb = lamb[mask]
        if np.size(a) > 1:
            a = a[mask]
        if np.size(gamma) > 1:
            gamma = gamma[mask]
        mask1 = abs_x[mask] <= (1 + gamma) * lamb
        mask2 = (abs_x[mask] > (1 + gamma) * lamb) * (abs_x[mask] <= a * lamb)
        mask3 = abs_x[mask] > a * lamb
        x_ = x[mask]
        prox_mask = (
            1.0 * sign_x[mask] * np.maximum(abs_x[mask] - lamb * gamma, 0) * mask1
            + ((a - 1) * x_ - sign_x[mask] * a * lamb * gamma) / (a - 1 - gamma) * mask2
            + x_ * mask3
        )

        prox_x[mask] = prox_mask
        return np.reshape(prox_x, np.shape(x))

    def __call__(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> float:
        self._check_size(x, 1)
        a = self.a
        lamb = self.lamb
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        # 1st branch
        abs_x = np.abs(1.0 * x)
        fun_x = lamb * abs_x

        # 2nd branch
        mask = np.logical_and(abs_x > lamb, abs_x <= a * lamb)
        fun_x[mask] = -(0.5 * (lamb**2 - 2 * a * lamb * abs_x + x**2) / (a - 1))[
            mask
        ]

        # 3rd branch
        mask = abs_x > a * lamb
        if np.size(lamb) > 1:
            lamb = lamb[mask]
        if np.size(a) > 1:
            a = a[mask]
        fun_x[mask] = 0.5 * (a + 1) * lamb**2

        return np.sum(fun_x)

    def _check_size(self, x, gamma):
        sz_x = np.size(x)
        if np.any(gamma <= 0):
            raise Exception(
                "'gamma' (or all of its components if it"
                + " is an array) must be strictly positive"
            )
        if (np.size(self.lamb) > 1) and (np.size(self.lamb) != sz_x):
            raise Exception("'lamb' must be either scalar or the same size as 'x'")
        if (np.size(self.a) > 1) and (np.size(self.a) != sz_x):
            raise Exception("'a' must be either scalar or the same size as 'x'")
        if (np.size(gamma) > 1) and (np.size(gamma) != sz_x):
            raise Exception("'gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-20-2022).

DEPENDENCIES:
     -'solver_cubic.py' - located in the folder 'utils'
     -'newton.py'  - located in the folder 'utils

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np
from proxop.utils.newton import newton_


class Renyi:
    r"""Compute the proximity operator and the evaluation of gamma*D.

    Where D is the function defined as:


                  /  (x - y)^alpha / y^{alpha-1}   if x >= 0 and y > 0
        D(x,y) = |   0                             if x=y=0
                 \  + inf                          otherwise

    'gamma' is the scale factor

    When the inputs are arrays, the outputs are computed element-wise.


    INPUTS
    ========
    x          - scalar or ND array
    y          - scalar or ND array with the same size as 'x'
    alpha      - > 1, scalar      [default: alpha=2 ]
    gamma      - scalar or ND array compatible with the blocks of 'y'
                [default: gamma=1]
    """

    def __init__(self, alpha: float = 2):

        if np.any(alpha <= 1) or np.size(alpha) > 1:
            raise ValueError("'alpha'  must be a scalar greter than 1")
        self.alpha = alpha

    def prox(
            self,
            x: np.ndarray,
            y: np.ndarray,
            gamma: Union[float, np.ndarray] = 1
    ) -> Tuple[np.ndarray, np.ndarray]:
        alpha = self.alpha
        if np.any(gamma <= 0):
            raise ValueError("'gamma'  must be strictly positive")
        if np.size(x) != np.size(y):
            raise ValueError("'x' and 'y' must have the same size")

        # scalar-like inputs handling
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
            y = np.reshape(y, (-1))

        # 2nd branch
        prox_p = np.zeros(np.shape(x))
        prox_q = np.maximum(0, y)

        # branch selection
        mask = (x > 0) * (
            gamma ** (1 / (alpha - 1)) * y
            < (1 - alpha) * (x / alpha) ** (1 + 1 / (alpha - 1))
        )

        xx = x[mask]
        yy = y[mask]

        if np.size(gamma) <= 1:
            gg = gamma
        else:
            gg = gamma[mask]

        # newton's method
        def fun_phi(t):
            return (
                xx / gg * t ** (alpha + 1)
                - yy / gg * t**alpha
                - alpha * t**2
                + 1
                - alpha
            )

        def der_fun_phi(t):
            return (
                (alpha + 1) * xx / gg * t**alpha
                - alpha * yy / gg * t ** (alpha - 1)
                - 2 * alpha * t
            )

        # root finding
        low = (alpha * gg / xx) ** (1 / alpha - 1)
        root = 100 / alpha + low
        root = newton_(fun_phi, fp=der_fun_phi, x0=root, low=low, high=np.inf)

        # 1st branch
        prox_p[mask] = xx - alpha * gg / root ** (alpha - 1)
        prox_q[mask] = yy + (alpha - 1) * alpha / root**alpha
        return tuple([prox_p, prox_q])

    def __call__(
            self,
            x: np.ndarray,
            y: np.ndarray
    ) -> float:
        if np.any(x < 0) or np.any(y < 0) or np.any((y == 0) * (x != 0)):
            return np.inf

        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
            y = np.reshape(y, (-1))
        alpha = self.alpha
        mask = y == 0
        res = x[mask] ** alpha / y[mask] ** (alpha - 1)

        return np.sum(res)



"""
Compute the proximity operator and the evaluation of multivariate functions.

Author: Mbaye DIONGUE (Jun 2022)
"""

from .AbsDiff import *
from .AffineBarrier import *
from .ChiSquare import *
from .Hellinger import *
from .HyperslabBarrier import *
from .Ialpha import *
from .InvPenaltyLogDet import *
from .Jeffrey import *
from .Kullback import *
from .L1Norm import *
from .L2BallBarrier import *
from .L2DownConv import *
from .L2Norm import *
from .L2Positive import *
from .L21columns import *
from .L21rows import *
from .Linf import *
from .Linf1 import *
from .LogDet import *
from .Max import *
from .NeumannEntropy import *
from .NuclearBall import *
from .NuclearBlocks import *
from .NuclearLogDet import *
from .NuclearNorm import *
from .NuclearNormRidge import *
from .PermutationInvariant import *
from .PersAbsDiff import *
from .PersHuber import *
from .PersHuber import *
from .PersSqrt import *
from .PersSquare import *
from .PersVapnik import *
from .Rank import *
from .RankRidge import *
from .Renyi import *
from .Schatten3_2Penalty import *
from .Schatten3Penalty import *
from .Schatten4_3Penalty import *
from .Schatten4Penalty import *
from .SchattenPenaltyLogDet import *
from .Spectral import *
from .SquaredFrobeniusNormLogDet import *
from .Vapnik import *


__all__ = [
    "AbsDiff",
    "AffineBarrier",
    "ChiSquare",
    "Hellinger",
    "HyperslabBarrier",
    "Ialpha",
    "InvPenaltyLogDet",
    "Jeffrey",
    "Kullback",
    "L1Norm",
    "L2BallBarrier",
    "L2DownConv",
    "L2Norm",
    "L2Positive",
    "L21columns",
    "L21rows",
    "Linf",
    "Linf1",
    "LogDet",
    "Max",
    "NeumannEntropy",
    "NuclearBall",
    "NuclearBlocks",
    "NuclearLogDet",
    "NuclearNorm",
    "NuclearNormRidge",
    "PermutationInvariant",
    "PersAbsDiff",
    "PersHuber",
    "PersHuber",
    "PersSqrt",
    "PersSquare",
    "PersVapnik",
    "Rank",
    "RankRidge",
    "Renyi",
    "Schatten3_2Penalty",
    "Schatten3Penalty",
    "Schatten4_3Penalty",
    "Schatten4Penalty",
    "SchattenPenaltyLogDet",
    "Spectral",
    "SquaredFrobeniusNormLogDet",
    "Vapnik"
]



"""
Version : 1.0 ( 06-22-2022).

DEPENDENCIES:
    -'newton.py' - located in the folder 'utils'
    -'Renyi.py'  - located in the folder 'multi'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np
from proxop.multi.Renyi import Renyi


class ChiSquare:
    r"""Compute the proximity operator and the evaluation of gamma*D.

    Where D is the  function defined as:


                  /  (x - y)^2 / y   if x >= 0 and y > 0
        D(x,y) = |   0               if x=y=0
                 \  + inf            otherwise

    'gamma' is the scale factor

    When the inputs are arrays, the outputs are computed element-wise

    INPUTS
    ========
    x        - scalar or ND array
    y        - scalar if 'x' is a scalar , ND array with the same size as 'x' otherwise
    gamma    - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    """

    def __init__(self):
        pass

    def prox(
            self,
            x: np.ndarray,
            y: np.ndarray,
            gamma: Union[float, np.ndarray] = 1
    ) -> Tuple[np.ndarray, np.ndarray]:
        self._check(x)
        return Renyi(alpha=2).prox(x + 2 * gamma, y - gamma, gamma)

    def __call__(self, x: np.ndarray, y: np.ndarray) -> float:
        if np.size(x) != np.size(y):
            raise ValueError("'x' and 'y' must have the same size")
        if np.any(x < 0) or np.any(y <= 0):
            return np.inf
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
            y = np.reshape(y, (-1))
        res = (x - y) ** 2 / y
        res[(x == 0) * (y == 0)] = 0
        return np.sum(self.gamma * res)

    def _check(self, x, gamma=1):
        if np.any(gamma <= 0):
            raise ValueError("'gamma'  must be strictly positive")
        if np.size(self.gamma) > 1 and np.size(self.gamma) != np.size(x):
            ValueError(
                "'gamma' must be positive scalars or positive ND arrays" +
                " with the same size as 'x'"
            )



"""
Version : 1.0 ( 06-23-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class BregAbsLog:
    r"""Compute the proximity operator prox_{gamma*f}^{phi}.

    Where the functions f and phi are defined as follows:

              f(x)=  | x - delta |

                      /  - log(u)     if u>0
             phi(u)= |
                     \   +inf         otherwise

    Note: The function phi is a Legendre type function and induces a Bregman distance

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    - When calling the functions f or phi (method 'phi') the result is computed
    element-wise SUM.
    So the command >>>BregAbsLog()(x) or >>>BregAbsLog().phi(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'.
    So, the command >>>BregAbsLog.prox(x)   will return an array with the
    same shape as 'x'

    INPUTS
    ========
    x       - scalar or ND array
    delta   - positive, scalar or ND array with the same size as 'x'[default: delta=1]
    gamma   - positive, scalar or ND array with the same size as 'x'[default: gamma=1]

    ========
     Examples
     ========

     Evaluate the  function f :

     >>> BregAbsLog(delta=0.001)(4)
     3.999
     >>> BregAbsLog(delta=0.001)( [-2, 3, 4] )
     8.999

     Evaluate The function  phi

     >> BregAbsLog(delta=0.001).phi(4)
     -1.3862943611198906

     Compute the proximity operator at a given point :

     >>> BregAbsLog(delta=0.1).prox( [-2, 3, 4] )
     array([2.  , 0.75, 0.8 ])
     >>> BregAbsLog(delta=0.1, gamma=2).prox( [-2, 3, 4] )
     array([0.66666667, 0.42857143, 0.44444444])
    """

    def __init__(self, delta: Union[float, np.ndarray]):
        if np.size(delta) > 1 and (not isinstance(delta, np.ndarray)):
            delta = np.array(delta)
        if np.any(delta <= 0):
            raise ValueError("'delta'  must be strictly positive")
        self.delta = delta

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        scale = gamma
        delta = self.delta
        self._check(x, gamma)

        # scalar-like inputs handling
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        # 1st branch
        prox_x = x / (1 + scale * x)

        # 2nd branch
        mask1 = x <= delta / (1 + scale * delta)
        eps = 1.51e-15
        mask = np.logical_and(mask1, (x >= delta / (1 - scale * delta + eps)))
        dd = delta
        if np.size(dd) > 1:
            dd = dd[mask]
        prox_x[mask] = dd

        # 3rd branch
        mask = np.logical_not(mask1)
        gg = scale
        if np.size(gg) > 1:
            gg = gg[mask]
        prox_x[mask] = x[mask] / (1 + gg * x[mask])

        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        return np.sum(np.abs(x - self.delta))

    def phi(self, u: np.ndarray) -> float:
        if np.size(u) > 1 and (not isinstance(u, np.ndarray)):
            u = np.array(u)
        if np.size(u) <= 1:
            u = np.reshape(u, (-1))
        res = np.zeros(np.shape(u))
        mask = u > 0
        res[mask] = -np.log(u[mask])
        res[u <= 0] = np.inf
        return np.sum(res)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError("'gamma'  must be strictly positive")
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")
        if (np.size(self.delta) > 1) and (np.size(self.delta) != np.size(x)):
            raise ValueError("delta' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-20-2022).

DEPENDENCIES:
    - 'L2Norm.py' located in the folder 'multi'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.multi.L2Norm import L2Norm


class L21rows:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the sum of euclidian norms of the rows of a matrix:

        f(x) = f(x) = \sum_{i=1}^N|\sum_{j=1}^M|X(i,j)|^2|^{\frac{1}{2}}

            where X = U*diag(s)*V.T \in R^{M*N}  (Singular Value decomposition)


     INPUTS
    ========
     x         -  (M,N) -array_like ( representing an M*N matrix )
     gamma     - positive, scalar or ND array compatible with the size of 'x'
                 [default: gamma=1]
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        return L2Norm(axis=1).prox(x, gamma)

    def __call__(self, x: np.ndarray) -> float:
        return L2Norm(axis=1)(x)



"""
Version : 1.0 ( 06-16-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np


class EpiSupport:
    r"""Compute the projection and the indicator of the epigraph of phi.

    Where phi is the function defined as:


                               /  * a * x   if x < 0
     phi(y)= sigma_[a,b](y) = |   0         if x = 0
                              \   b * x     otherwise


    When the inputs are arrays, the outputs are computed element-wise
     INPUTS
    ========
     y    - ND array
     ksi  - ND array with the same size as 'y'
     a    - negative, scalar or ND array with the same size as x
     b    - positive, scalar or ND array with the same size as x
    """

    def __init__(
        self,
        ksi: np.ndarray,
        a: Union[float, np.ndarray] = -1.0,
        b: Union[float, np.ndarray] = 1.0
    ):
        if np.any(a > 0):
            raise ValueError("'a' (or all of its components ) must be negative")
        if np.any(b < 0):
            raise ValueError("'b' (or all of its components ) must be positive")
        if np.size(a) > 1 and (np.size(a) != np.size(ksi)):
            raise ValueError(" 'a' must be a scalar or have the same size as 'ksi'")
        if np.size(b) > 1 and (np.size(b) != np.size(ksi)):
            raise ValueError(" 'b' must be a scalar or have the same size as 'ksi'")
        if np.size(ksi) <= 1:
            ksi = np.reshape(ksi, (-1))
        self.a = a
        self.b = b
        self.ksi = ksi

    # proximal operator (i.e. the projection on the constraint set)
    def prox(self, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        self._check(y)
        ksi = self.ksi
        a = self.a
        b = self.b
        if np.size(y) <= 1:
            y = np.reshape(y, (-1))

        # 4th branch
        prox_p = np.zeros(np.size(y))
        prox_t = np.zeros(np.size(y))

        # 3rd branch
        mask = np.logical_and(a * y > ksi, -y / a <= ksi)
        pp = (y + a * ksi) / (1 + a**2)
        tt = a * pp
        prox_p[mask] = pp[mask]
        prox_t[mask] = tt[mask]

        # 2nd branch
        mask = np.logical_and(b * y > ksi, -y / b <= ksi)
        pp = (y + b * ksi) / (1 + b**2)
        tt = b * pp
        prox_p[mask] = pp[mask]
        prox_t[mask] = tt[mask]

        # 1st branch
        mask = np.logical_and(a * y <= ksi, b * y <= ksi)
        prox_p[mask] = y[mask]
        prox_t[mask] = ksi[mask]
        return tuple([prox_p, prox_t])

    # indicator of the constraint set
    def __call__(self, y) -> float:
        if np.all(self.a * y <= self.ksi) and np.all(self.b * y <= self.ksi):
            return 0
        return np.inf

    def _check(self, y):
        if np.size(y) != np.size(self.ksi):
            raise ValueError(" 'y' must have the same size as 'ksi'")



"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class Berhu:
    r"""Compute the proximity operator and the evaluation of the function f.

    Where f is the function defined as:

                  / ( x^2 + w^2 ) / ( 2* W)     if |x| > w
           f(x)= |
                 \  |w|                         otherwise

                 with w>0


    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function the result
    is computed element-wise SUM. So the command >>>Berhu()(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same shape
    as the input 'x'.
    So, the command >>>Berhu().prox(x)   will return an array with the same
    shape as 'x'

     INPUTS
    ========
     x     - ND array
     w     - positive, scalar or ND array with the same size as 'x'

     ========
     Examples
     ========

     Evaluate the 'direct' function:

     >>> Bh=Berhu(w=2)
     >>> Bh(3)
      3.25
     >> Bh([-1, 2, 3])
     7.25

     Compute the proximity operator at a given point :

     >>> Bh.prox( 3)
     array([2])
     >>> Bh.prox([ -3., 1., 6.])
     array([-2.,  0.,  4.])
    """

    def __init__(
            self,
            w: Union[float, np.ndarray]
    ):
        if np.size(w) > 1 and (not isinstance(w, np.ndarray)):
            w = np.array(w)
        if np.any(w <= 0):
            raise ValueError(
                "'w' (or all of its components if it is an array) must "
                + "be strictly positive"
            )
        self.w = w

    def prox(self, x: np.ndarray) -> np.ndarray:
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        self._check(x, self.w)
        # preliminaries
        abs_x = np.abs(x)
        sign_x = np.sign(x)
        w = self.w

        # 2nd branch
        prox_x = x - sign_x

        # 1st branch
        mask = abs_x > w + 1

        if np.size(w) > 1:
            prox_x[mask] = w[mask] * x[mask] / (1 + w[mask])
        else:
            prox_x[mask] = w * x[mask] / (w + 1)

        # third branch
        mask2 = abs_x <= 1
        prox_x[mask2] = 0

        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        self._check(x, self.w)
        abs_x = np.abs(x)
        w = self.w

        result = np.abs(w) * np.ones(np.shape(x))
        mask = abs_x > w

        if np.size(w) > 1:
            result[mask] = (w[mask] ** 2 + x[mask] ** 2) / (2 * w[mask])

        else:
            result[mask] = (w**2 + x[mask] ** 2) / (2 * w)

        return np.sum(result)

    def _check(self, x, w):
        if (np.size(w) > 1) and (np.size(w) != np.size(x)):
            raise ValueError("'w' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-8-2022).

Author  : Mbaye DIONGUE

DEPENDENCIES:
   -'Max.py' in the folder 'multi'

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Optional
import numpy as np
from proxop.multi.Max import Max


class Simplex:
    """Compute the projection and the indicator of the simplex.

    Recall: every vector X belonging to the simplex verifies:

                    x >= 0 and  (1,..., 1).T * X = eta


     where (1, ..., 1) is a ND array with all components equal to one,
     and (1,..., 1).T its transpose

     INPUTS
    ========
     x    - ND array
     eta  - positive, scalar or ND array compatible with the blocks of 'x'
     axis - int or None, direction of block-wise processing [DEFAULT: axis=None]
            When the input 'x' is an array, the computation can vary as follows:
            - axis = None --> 'x' is processed as a single vector
            - axis >= 0 --> 'x' is processed block-wise along the specified axis
              (axis=0 -> rows, axis=1 -> columns etc.).
    """

    def __init__(
            self,
            eta: Union[float, np.ndarray],
            axis: Optional[int] = None
    ):
        if np.any(eta <= 0):
            raise Exception(
                "'eta' (or all of its components if it is an array) must be positive"
            )
        self.eta = eta
        self.axis = axis

    # proximal operator (i.e projection on the simplex)
    def prox(self, x: np.ndarray) -> np.ndarray:
        return x - Max(axis=self.axis).prox(x, gamma=self.eta)

    # indicator of the simplex
    def __call__(self, x: np.ndarray) -> float:
        """
        Indicate if the input 'x' is in the constraint set or not
        Parameters
        ----------
        x : np.ndarray

        Returns
        -------
        0      if the input 'x' is in the set
        +inf   otherwise
        """
        scalar_prod = np.sum(x, axis=self.axis)
        if np.all(x >= 0) and np.all(scalar_prod==self.eta):
            return 0
        return np.inf



"""
Version : 1.0 (06-18-2022).

DEPENDENCIES:
     'solver_cubic.py' in the folder 'utils'

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.solver_cubic import solver_cubic


class BurgLogSum:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

      where the function f is defined as:

                       /   -log(x) + w*log(delta+x)   if x>0
               f(x) = |
                      \   +inf                      otherwise

    with w>0 and delta>0
    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    delta -positive, scalar or ND array with the same size as 'x'
            [default: delata=1e-10]
    w     - positive, scalar or ND array with the same size as 'x' [default: w=1]


    Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>BurgLogSum(w=2)(x) will
    return a scalar even if x is a vector:

    >>> BurgLogSum(w=2)(np.array([10, 20, 30]))
    8.699514748246859

    But as expected, >>>BurgLogSum(w=2).prox(x)
    will return a vector with the same size as x:

    >>> BurgLogSum(w=2).prox(np.array([10, 20, 30]))
    array([ 9.89897949, 19.94987437, 29.96662955])
    """

    def __init__(
        self,
        w: Union[np.ndarray, float] = 1.0,
        delta: Union[np.ndarray, float] = 1e-10
    ):
        if np.any(delta <= 0):
            raise ValueError("'delta' (or all of its elements)" +
                             " must be strictly positive")
        if np.any(w <= 0):
            raise ValueError("'w' (or all of its elements)" +
                             " must be strictly positive")
        self.delta = delta
        self.w = w

    def prox(self, x: np.ndarray, gamma: Union[np.ndarray, float] = 1.0) -> np.ndarray:
        self._check(x, gamma)
        delta = self.delta
        w = self.w
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        # Solve the cubic equation
        prox_x, p2, p3 = solver_cubic(
            1, delta - x, w * gamma - delta * x - gamma, -gamma * delta
        )

        def f(t):
            res = np.inf * np.ones_like(t)
            mask_ = t > 0
            w_ = w
            delta_ = delta
            if np.size(delta) > 1:
                delta_ = delta[mask_[0]]
            if np.size(w) > 1:
                w_ = w[mask_[0]]
            res[mask_] = -np.log(t[mask_]) + w_ * np.log(delta_ + t[mask_])
            return 0.5 * np.abs(x - t) ** 2 + gamma * res

        mask = np.logical_and(
            np.isreal(p2) * (np.real(p2) > 0), f(np.real(p2)) < f(prox_x)
        )

        prox_x[mask] = np.real(p2[mask])
        mask = np.logical_and(
            np.isreal(p2) * (np.real(p3) > 0), f(np.real(p3)) < f(prox_x)
        )
        prox_x[mask] = np.real(p3[mask])

        return np.reshape(prox_x, np.shape(x))

    def __call__(self, x: np.ndarray) -> float:
        if np.any(x <= 0):
            return np.inf
        fun_x = -np.log(x) + self.w * np.log(self.delta + x)
        return np.sum(fun_x)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError("'gamma' (or all of its elements)" +
                             " must be strictly positive")
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("'gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 (06-18-2022).

DEPENDENCIES:
     'newton.py' in the folder 'utils'

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.newton import newton_


class EntropyCauchy:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

      where the function f is defined as:

                       /   xlog(x)+ w*log( delta + x^2 )    if x>0
               f(x) = |    w*log(delta)                      if x=0
                      \   +inf                               otherwise

    with w>0 and delta>0

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     -ND array
    w     -positive, scalar or ND array with the same size as 'x' [default: w=1]
    delta -positive, scalar or ND array with the same size as 'x'[default: delata=1e-10]
    gamma -positive, scalar or ND array with the same size as 'x' [default: gamma=1]


     Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>EntropyCauchy(delta=0.01, w=2)(x)
    will
    return a scalar even if x is a vector:

    >>> EntropyCauchy(delta=0.01, w=2)(np.array([10, 20, 30]))
    219.77474905520018

    But as expected, >>>EntropyCauchy(delta=0.01, w=2).prox(x)
    will return a vector with the same size as x:

    >>> EntropyCauchy(delta=0.01, w=2).prox(np.array([10, 20, 30]))
    array([ 7.44509635, 16.92485178, 26.55998001])
    """

    def __init__(
        self,
        w: Union[np.ndarray, float] = 1.0,
        delta: Union[np.ndarray, float] = 1e-10
    ):

        if np.any(delta <= 0):
            raise ValueError(
                "'delta'(or all of its elements)" + " must be strictly positive"
            )
        if np.any(w <= 0):
            raise ValueError(
                "'w' (or all of its elements)" + " must be strictly positive"
            )
        self.delta = delta
        self.w = w

    def prox(self, x: np.ndarray, gamma: Union[np.ndarray, float] = 1.0) -> np.ndarray:
        self._check_size(x, gamma)
        delta = self.delta
        w = self.w

        def f(t):
            return (
                t**3
                + (delta - x) * t**2
                + (delta + 2 * gamma * w) * t
                + gamma * delta * np.log(t)
                + gamma * t**2 * np.log(t)
                + delta * (gamma - x)
            )

        def df(t):
            return (
                3 * t**2
                + 2 * (gamma - x) * t
                + (delta + 2 * gamma * w)
                + gamma * delta / t
                + gamma * t * (2 * np.log(t) + 1)
            )

        abs_x = np.abs(x)
        prox_x = np.maximum(1, abs_x)
        prox_x = newton_(f, df, prox_x, low=0, high=10 * abs_x)
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        self._check_size(x, 1)
        if np.any(x < 0):
            return np.inf
        res = np.zeros(np.shape(x))
        mask = x > 0
        w = self.w
        delta = self.delta
        if np.size(delta) > 1:
            delta = delta[mask]
        if np.size(w) > 1:
            w = w[mask]
        res[mask] = x[mask] * np.log(x[mask]) + w * np.log(delta + x[mask] ** 2)
        mask = x == 0
        if np.size(self.delta) > 1:
            delta = self.delta[mask]
        if np.size(self.w) > 1:
            w = self.w[mask]
        res[mask] = w * np.log(delta)
        return np.sum(res)

    def fun(self, x: np.ndarray) -> np.ndarray:
        res = np.inf * np.ones_like(1.0 * x)
        mask = x > 0
        w = self.w
        delta = self.delta
        if np.size(delta) > 1:
            delta = delta[mask]
        if np.size(w) > 1:
            w = w[mask]
        res[mask] = x[mask] * np.log(x[mask]) + w * np.log(delta + x[mask] ** 2)
        mask = x == 0
        if np.size(self.delta) > 1:
            delta = self.delta[mask]
        if np.size(self.w) > 1:
            w = self.w[mask]
        res[mask] = w * np.log(delta)
        return res

    def _check_size(self, x, gamma):
        sz_x = np.size(x)
        if (np.size(self.w) > 1) and (np.size(self.w) != sz_x):
            raise ValueError("'w' must be either scalar or the same size as 'x'")
        if (np.size(self.delta) > 1) and (np.size(self.delta) != sz_x):
            raise ValueError("'delta' must be either scalar or the same size as 'x'")
        if (np.size(gamma) > 1) and (np.size(gamma) != sz_x):
            raise ValueError("'gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-21-2022).

DEPENDENCIES:
     - 'prox_svd.py' located in the folder 'utils'
     - 'prox_svd.py' located in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from proxop.utils.prox_svd import prox_svd
from proxop.utils.fun_svd import fun_svd
import numpy as np


class LogDet:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the logarithm of determinant function:

               /-log( det(X) ) = -log( prod(s))  if X is a symmetric positive
         f(x)=|                                  definite matrix
              \ + inf                            otherwise

    where
           * det(X) is the determinant of the matrix X

           * X = U*diag(s)*V.T \in R^{M*N}  the Singular Value decomposition of X

          * 'gamma*tau' is the scale factor

     INPUTS
    ========
     x               -  (M,N) -array_like ( representing an M*N matrix)
     gamma           - positive scalar  [default: gamma=1]
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray, gamma: float = 1) -> np.ndarray:
        self._check(x, gamma)
        is_hermitian = False
        if np.allclose(x, np.transpose(x)):
            is_hermitian = True

        def prox_phi(s, gam):
            return 0.5 * (s + np.sqrt(s**2 + 4 * gam))

        return prox_svd(x, gamma, prox_phi, hermitian=is_hermitian)

    def __call__(self, x: np.ndarray) -> float:
        self._check(x)
        tol = 1e-20
        # Check if the matrix is symmetric
        if not np.allclose(x, np.transpose(x)):
            return np.inf

        def fun_phi(s):
            if np.any(s <= tol):
                return np.inf
            return -np.log(np.prod(s))

        return fun_svd(x, 1, fun_phi)

    def _check(self, x, gamma=1):
        if np.any(gamma <= 0) or np.size(gamma) > 1:
            raise Exception("'gamma'  must be a strictly positive scalar")
        if len(np.shape(x)) != 2:
            raise ValueError(
                "'x' must be an (M,N) -array_like ( representing a M*N matrix )"
            )



"""
Version : 1.0 ( 06-22-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np


class AbsDiff:
    r"""Compute the proximity operator and the evaluation of gamma*D.

    Where D is the  function defined as:

                  /  |x-y|        if x >= 0 and y >= 0
        D(x,y) = |
                 \  + inf         otherwise

    'gamma' is the scale factor

    When the inputs are arrays, the outputs are computed element-wise

    ========
     INPUTS
    ========
    x     - ND array
    y     - ND array with the same size as 'x'
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    """

    def __init__(self):
        pass

    def prox(
            self,
            x: np.ndarray, y: np.ndarray,
            gamma: Union[float, np.ndarray] = 1
    ) -> Tuple[np.ndarray, np.ndarray]:
        if np.size(x) != np.size(y):
            raise ValueError("'x' and 'y' must have the same size")
        scale = gamma
        self._check(x, gamma)

        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
            y = np.reshape(y, (-1))
        sz = np.shape(x)

        # 4th branch
        prox_p = np.zeros(sz)
        prox_q = np.zeros(sz)

        # 3rd branch
        mask = (y > scale) * (x <= -scale)
        yy = y - scale
        prox_q[mask] = yy[mask]

        # 2nd branch
        mask = (x > scale) * (y <= -scale)
        xx = x - scale
        prox_p[mask] = xx[mask]

        # 1st branch
        t = np.sign(x - y) * np.maximum(0, np.abs(x - y) - 2 * scale)
        mask = np.abs(t) < x + y
        xy = x[mask] + y[mask]
        tt = t[mask]
        prox_p[mask] = 0.5 * (xy + tt)
        prox_q[mask] = 0.5 * (xy - tt)

        return tuple([prox_p, prox_q])

    def __call__(self, x: np.ndarray, y: np.ndarray) -> float:
        if np.any(x < 0) or np.any(y < 0):
            return np.inf
        return np.sum(np.abs(x - y))

    @staticmethod
    def _check(x, gamma=1):
        if np.any(gamma <= 0):
            raise ValueError("'gamma'  must be strictly positive")
        if np.size(gamma) > 1 and np.size(gamma) != np.size(x):
            ValueError(
                "'gamma' must be positive scalars or positive ND arrays" +
                " with the same size as 'x'"
            )
